from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal import BaseAsset
from mesido.pycml.component_library.milp.heat.heat_four_port import HeatFourPort

from numpy import nan


class HeatExchanger(HeatFourPort, BaseAsset):
    """
    The heat exchanger component is used to model the exchange of thermal power between two
    hydraulically decoupled systems. A constant efficiency is used to model milp losses and a
    maximum power is set on the primary side to model physical constraints on the amount of heat
    transfer.

    The heat to discharge constraints are set in the HeatMixin. The primary side is modelled as a
    demand, meaning it consumes energy from the primary network and gives it to the secondary side,
    where the secondary side acts like a source to the secondary network. This also means that heat
    can only flow from primary to secondary.

    To avoid unphysical heat transfer the HeatPhysicsMixin sets constraints on the temperatures on
    both sides in the case of varying temperature. We also allow a heat_exchanger to be disabled on
    certain time-steps to then allow these temperature constraints to be also disabled.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(),
                modifiers,
            ),
        )

        self.component_type = "heat_exchanger"
        self.efficiency = nan

        self.nominal = (
            self.Secondary.Q_nominal * self.Secondary.rho * self.Secondary.cp * self.Secondary.dT
        )

        self.price = nan
        self.minimum_pressure_drop = 1.0e5  # 1 bar of pressure drop
        self.pump_efficiency = 0.5

        # Assumption: heat in/out and added is nonnegative

        self.add_variable(Variable, "Primary_heat", min=0.0)
        self.add_variable(Variable, "Secondary_heat", min=0.0)
        self.add_variable(Variable, "Heat_flow", nominal=self.nominal)
        self.add_variable(Variable, "dH_prim")
        self.add_variable(Variable, "dH_sec")

        # Hydraulically decoupled so Heads remain the same
        self.add_equation(self.dH_prim - (self.Primary.HeatOut.H - self.Primary.HeatIn.H))
        self.add_equation(
            (
                self.minimum_pressure_drop * self.Primary.Q
                - (self.Primary.HeatIn.Hydraulic_power - self.Primary.HeatOut.Hydraulic_power)
            )
            / (self.Primary.Q_nominal * self.Primary.nominal_pressure)
        )
        self.add_equation(self.dH_sec - (self.Secondary.HeatOut.H - self.Secondary.HeatIn.H))
        self.add_equation(
            (
                self.Pump_power
                - (self.Secondary.HeatOut.Hydraulic_power - self.Secondary.HeatIn.Hydraulic_power)
            )
            / (self.Secondary.Q_nominal * self.Secondary.nominal_pressure)
        )

        self.add_equation(
            ((self.Primary_heat * self.efficiency - self.Secondary_heat) / self.nominal)
        )

        self.add_equation(
            (
                (self.Primary_heat - (self.Primary.HeatIn.Heat - self.Primary.HeatOut.Heat))
                / self.nominal
            )
        )
        self.add_equation(
            (
                (self.Secondary_heat - (self.Secondary.HeatOut.Heat - self.Secondary.HeatIn.Heat))
                / self.nominal
            )
        )
        self.add_equation((self.Heat_flow - self.Secondary_heat) / self.nominal)



import os
import time

import casadi as ca

from mesido.esdl.esdl_additional_vars_mixin import ESDLAdditionalVarsMixin
from mesido.esdl.esdl_mixin import ESDLMixin
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.techno_economic_mixin import TechnoEconomicMixin
from mesido.workflows.io.write_output import ScenarioOutput

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import SinglePassGoalProgrammingMixin
from rtctools.util import run_optimization_problem


class MaxRevenue(Goal):

    priority = 1

    order = 1

    def __init__(self, asset_name: str):
        """
        The constructor of the goal.

        Parameters
        ----------
        source : string of the source name that is going to be minimized
        """

        self.asset_name = asset_name

    def function(
        self, optimization_problem: CollocatedIntegratedOptimizationProblem, ensemble_member: int
    ) -> ca.MX:
        """
        This function returns the state variable to be minimized.

        Parameters
        ----------
        optimization_problem : The optimization class containing the variables'.
        ensemble_member : the ensemble member.

        Returns
        -------
        The negative revenue var of the optimization problem.
        """
        return -optimization_problem.extra_variable(f"{self.asset_name}__revenue", ensemble_member)


class MinCost(Goal):

    priority = 1

    order = 1

    def __init__(self, asset_name: str):
        # self.target_max = 0.0
        # self.function_range = (0.0, 1.0e9)
        # self.function_nominal = 1.0e7

        self.asset_name = asset_name

    def function(
        self, optimization_problem: CollocatedIntegratedOptimizationProblem, ensemble_member: int
    ) -> ca.MX:

        return optimization_problem.extra_variable(
            f"{self.asset_name}__fixed_operational_cost", ensemble_member
        ) + optimization_problem.extra_variable(
            f"{self.asset_name}__variable_operational_cost", ensemble_member
        )


class EmergeWorkFlow(
    ScenarioOutput,
    ESDLAdditionalVarsMixin,
    TechnoEconomicMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This problem class is for the absolute heat tests. Meaning that this problem class
    is applied to an esdl where there is no dedicated supply or return line. For this test case
    we just match heating demand (_GoalsAndOptions) and minimize the energy production to have a
    representative result.
    """

    # TODO: this workflow is still work in progress and this part of code still needs to be
    #  finalised

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.gas_network_settings["head_loss_option"] = HeadLossOption.NO_HEADLOSS

    def goals(self):

        goals = super().goals().copy()

        for asset_name in [
            *self.energy_system_components.get("electricity_demand", []),
            *self.energy_system_components.get("gas_demand", []),
        ]:
            goals.append(MaxRevenue(asset_name))
            goals.append(MinCost(asset_name))

        for asset_name in [
            *self.energy_system_components.get("electricity_source", []),
            *self.energy_system_components.get("gas_tank_storage", []),
            *self.energy_system_components.get("electricity_storage", []),
            *self.energy_system_components.get("electrolyzer", []),
            *self.energy_system_components.get("heat_pump_elec", []),
        ]:
            goals.append(MinCost(asset_name))

        return goals

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        for gs in self.energy_system_components.get("gas_tank_storage", []):
            canonical, sign = self.alias_relation.canonical_signed(f"{gs}.Stored_gas_mass")
            storage_t0 = sign * self.state_vector(canonical, ensemble_member)[0]
            constraints.append((storage_t0, 0.0, 0.0))
            canonical, sign = self.alias_relation.canonical_signed(f"{gs}.Gas_tank_flow")
            gas_flow_t0 = sign * self.state_vector(canonical, ensemble_member)[0]
            constraints.append((gas_flow_t0, 0.0, 0.0))

        for es in self.energy_system_components.get("electricity_storage", []):
            canonical, sign = self.alias_relation.canonical_signed(f"{es}.Stored_electricity")
            storage_t0 = sign * self.state_vector(canonical, ensemble_member)[0]
            constraints.append((storage_t0, 0.0, 0.0))

        return constraints

    def solver_options(self):
        """
        This function does not add anything at the moment but during debugging we use this.

        Returns
        -------
        solver options dict
        """
        options = super().solver_options()
        options["solver"] = "highs"
        return options

    def times(self, variable=None):
        return super().times(variable)[:25]

    def energy_system_options(self):
        """
        This function does not add anything at the moment but during debugging we use this.

        Returns
        -------
        Options dict for the physics modelling
        """
        options = super().energy_system_options()
        options["minimum_velocity"] = 0.0
        options["heat_loss_disconnected_pipe"] = False
        options["neglect_pipe_heat_losses"] = False
        options["include_asset_is_switched_on"] = True
        options["include_electric_cable_power_loss"] = False
        return options

    def post(self):
        # In case the solver fails, we do not get in priority_completed(). We
        # append this last priority's statistics here in post().
        # TODO: check if we still need this small part of code below
        success, _ = self.solver_success(self.solver_stats, False)
        # if not success:
        #     time_taken = time.time() - self.__priority_timer
        #     self._priorities_output.append(
        #         (
        #             self.__priority,
        #             time_taken,
        #             False,
        #             self.objective_value,
        #             self.solver_stats,
        #         )
        #     )

        super().post()

        # Optimized ESDL
        self._write_updated_esdl(self._ESDLMixin__energy_system_handler.energy_system)

        self._save_json = False

        if os.path.exists(self.output_folder) and self._save_json:
            self._write_json_output()

        results = self.extract_results()

        for _type, assets in self.energy_system_components.items():
            for asset in assets:
                print("----------------------------------")
                print(f"{asset} financials:")
                try:
                    print(f"revenue of {asset} in MEUR/day: ", results[f"{asset}__revenue"] / 1e6)
                except KeyError:
                    print(f"{asset} does not have a revenue")
                    pass
                try:
                    print(
                        f"fixed operational costs of {asset} in MEUR/yr : ",
                        results[f"{asset}__fixed_operational_cost"] / 1e6,
                    )
                    print(
                        f"variable operational costs of {asset} : ",
                        results[f"{asset}__variable_operational_cost"] / 1e6,
                    )  # not yet all included in financialmixin
                    print(f"max size of {asset} : ", results[f"{asset}__max_size"])
                except KeyError:
                    print(f"{asset} does not have a costs")
                    pass


if __name__ == "__main__":

    tic = time.time()
    elect = run_optimization_problem(
        EmergeWorkFlow,
        esdl_file_name="emerge_solar_battery.esdl",
        esdl_parser=ESDLFileParser,
        profile_reader=ProfileReaderFromFile,
        input_timeseries_file="timeseries_with_PV.csv",
    )
    print(time.time() - tic)
    results = elect.extract_results()
    a = 1



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.util import run_esdl_mesido_optimization

import numpy as np


class TestMILPGasSourceSink(TestCase):
    def test_source_sink(self):
        """
        Test case for a network consisting out of a source, pipes and a sink

        Checks:
        - That flow is maintained.
        - That the head drops over the pipe.

        """
        import models.unit_cases_gas.source_sink.src.run_source_sink as example
        from models.unit_cases_gas.source_sink.src.run_source_sink import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        # Added for case where head loss is modelled via DW
        class TestSourceSink(GasProblem):
            def energy_system_options(self):
                options = super().energy_system_options()
                self.heat_network_settings["head_loss_option"] = (
                    HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                )
                self.heat_network_settings["n_linearization_lines"] = 5
                self.heat_network_settings["minimize_head_losses"] = True

                self.heat_network_settings["pipe_maximum_pressure"] = 100.0  # [bar]
                self.heat_network_settings["pipe_minimum_pressure"] = 0.0
                return options

        soltion = run_esdl_mesido_optimization(
            GasProblem,
            base_folder=base_folder,
            esdl_file_name="source_sink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = soltion.extract_results()

        # Test if mass conserved
        np.testing.assert_allclose(
            results["GasProducer_0876.GasOut.Q"], results["GasDemand_a2d8.GasIn.Q"]
        )

        # Test if head is going down
        np.testing.assert_array_less(results["Pipe_4abc.GasOut.H"], results["Pipe_4abc.GasIn.H"])


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestMILPGasSourceSink()
    a.test_source_sink()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import sys
from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.workflows import run_end_scenario_sizing


class TestUpdatedESDL(TestCase):

    def test_updated_esdl(self):
        """
        Check that the updated ESDL resulting from optmizing a network, is correct by using the
        PoCTutorial and the Grow_workflow. This is done for the actual esdl file and the esdl
        string created by MESIDO. Both these resulting optimized energy systems should be identical
        and it is only the MESIDO esdl saving method that differs.

        """

        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.PoCTutorial.src.run_grow_tutorial
        from examples.PoCTutorial.src.run_grow_tutorial import EndScenarioSizingStagedHighs

        base_folder = (
            Path(examples.PoCTutorial.src.run_grow_tutorial.__file__).resolve().parent.parent
        )

        problem = run_end_scenario_sizing(
            EndScenarioSizingStagedHighs,
            base_folder=base_folder,
            esdl_file_name="PoC Tutorial.esdl",
            esdl_parser=ESDLFileParser,
        )

        # Save optimized esdl string
        optimized_esdl_string = problem.optimized_esdl_string
        file = open(
            Path.joinpath(base_folder, "model", "PoC Tutorial_GrowOptimized_esdl_string.esdl"), "w"
        )
        file.write(optimized_esdl_string)
        file.close()


if __name__ == "__main__":
    import time

    start_time = time.time()

    a = TestUpdatedESDL()
    a.test_updated_esdl()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

import mesido._darcy_weisbach as darcy_weisbach
from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.network_common import NetworkSettings
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test


class TestHeadLoss(TestCase):
    """
    Test case for a heat network and a gas network consisting out of a source, pipe(s) and a sink
    """

    def test_heat_network_head_loss(self):
        """
        Heat network: test the piecewise linear equality and inequality constraints of the head loss
        approximation.

        Checks:
        - That the head_loss() function does return the expected theoretical dH at a data point
        in the middle of the 1st line segment (dH curve is approximated with 5 linear lines)
        - That the head_loss() function does return a value smaller than a manual linearly
        approximated dH at a data point in the middle of the 1st line segment (dH curve is
        approximated with 5 linear lines)
        - That for the dH value approximated by the code is conservative, in other word greater
        than the theoretical value
        - That the pump power is conservative
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        for head_loss_option_setting in [
            HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY,
            HeadLossOption.LINEARIZED_N_LINES_EQUALITY,
        ]:
            # Added for case where head loss is modelled via DW
            class SourcePipeSinkDW(SourcePipeSink):
                # Do not delete: this is used to manualy check writing out of profile data
                # def post(self):
                #     super().post()
                #     self._write_updated_esdl(
                #       self._ESDLMixin__energy_system_handler.energy_system, optimizer_sim=True
                #   )

                def energy_system_options(self):
                    options = super().energy_system_options()

                    nonlocal head_loss_option_setting
                    head_loss_option_setting = head_loss_option_setting

                    if (
                        head_loss_option_setting
                        == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
                    ):
                        self.gas_network_settings["head_loss_option"] = (
                            HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                        )
                        self.gas_network_settings["n_linearization_lines"] = 5
                        self.heat_network_settings["minimize_head_losses"] = True
                    elif head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                        self.heat_network_settings["head_loss_option"] = (
                            HeadLossOption.LINEARIZED_N_LINES_EQUALITY
                        )
                        self.heat_network_settings["minimize_head_losses"] = True
                        self.heat_network_settings["minimum_velocity"] = 1.0e-6

                    return options

            # Do not delete kwargs: this is used to manualy check writing out of profile data
            kwargs = {
                "write_result_db_profiles": False,
                "influxdb_host": "localhost",
                "influxdb_port": 8086,
                "influxdb_username": None,
                "influxdb_password": None,
                "influxdb_ssl": False,
                "influxdb_verify_ssl": False,
            }

            solution = run_esdl_mesido_optimization(
                SourcePipeSinkDW,
                base_folder=base_folder,
                esdl_file_name="sourcesink.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries_import.csv",
                **kwargs,
            )
            results = solution.extract_results()

            pipes = ["Pipe1"]
            for itime in range(len(results[f"{pipes[0]}.dH"])):
                v_max = solution.heat_network_settings["maximum_velocity"]
                pipe_diameter = solution.parameters(0)[f"{pipes[0]}.diameter"]
                pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
                temperature = solution.parameters(0)[f"{pipes[0]}.temperature"]
                pipe_length = solution.parameters(0)[f"{pipes[0]}.length"]
                v_points = np.linspace(
                    0.0,
                    v_max,
                    solution.heat_network_settings["n_linearization_lines"] + 1,
                )
                v_inspect = (
                    results[f"{pipes[0]}.Q"][itime] / solution.parameters(0)[f"{pipes[0]}.area"]
                )
                idx = []
                linearized_idx = []
                idx.append(
                    (results["Pipe1.Q"][itime] / solution.parameters(0)["Pipe1.area"]) >= v_points
                )
                idx.append(
                    (results["Pipe1.Q"][itime] / solution.parameters(0)["Pipe1.area"]) < v_points
                )
                linearized_idx.append(np.where(idx[0])[0][-1])
                linearized_idx.append(np.where(idx[1])[0][0])

                # Theoretical head loss calc, dH =
                # friction_factor * 8 * pipe_length * volumetric_flow^2 / ( pipe_diameter^5 * g *
                # pi^2)
                dh_theory = (
                    darcy_weisbach.friction_factor(
                        v_inspect,
                        pipe_diameter,
                        pipe_wall_roughness,
                        temperature,
                    )
                    * 8.0
                    * pipe_length
                    * (v_inspect * np.pi * pipe_diameter**2 / 4.0) ** 2
                    / (pipe_diameter**5 * GRAVITATIONAL_CONSTANT * np.pi**2)
                )
                # Approximate dH [m] vs Q [m3/s] with a linear line between between v_points
                # dH_manual_linear = a*Q + b
                # Then use this linear function to calculate the head loss
                delta_dh_theory = darcy_weisbach.head_loss(
                    v_points[linearized_idx[1]],
                    pipe_diameter,
                    pipe_length,
                    pipe_wall_roughness,
                    temperature,
                ) - darcy_weisbach.head_loss(
                    v_points[linearized_idx[0]],
                    pipe_diameter,
                    pipe_length,
                    pipe_wall_roughness,
                    temperature,
                )

                delta_volumetric_flow = (
                    v_points[linearized_idx[1]] * np.pi * pipe_diameter**2 / 4.0
                ) - (v_points[linearized_idx[0]] * np.pi * pipe_diameter**2 / 4.0)

                a = delta_dh_theory / delta_volumetric_flow
                b = delta_dh_theory - a * delta_volumetric_flow
                dh_manual_linear = a * (v_inspect * np.pi * pipe_diameter**2 / 4.0) + b

                dh_milp_head_loss_function = darcy_weisbach.head_loss(
                    v_inspect, pipe_diameter, pipe_length, pipe_wall_roughness, temperature
                )

                np.testing.assert_allclose(dh_theory, dh_milp_head_loss_function)
                np.testing.assert_array_less(dh_milp_head_loss_function, dh_manual_linear)

                if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY:
                    np.testing.assert_array_less(
                        dh_manual_linear, -results[f"{pipes[0]}.dH"][itime] + 1e-6
                    )
                elif head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                    np.testing.assert_allclose(
                        -results[f"{pipes[0]}.dH"][itime], dh_manual_linear, rtol=1e-5, atol=1e-7
                    )

            for pipe in pipes:
                velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
                for ii in range(len(results[f"{pipe}.dH"])):
                    np.testing.assert_array_less(
                        darcy_weisbach.head_loss(
                            velocities[ii],
                            pipe_diameter,
                            pipe_length,
                            pipe_wall_roughness,
                            temperature,
                        ),
                        -results[f"{pipe}.dH"][ii],
                    )

            pump_power = results["source.Pump_power"]
            pump_power_post_process = (
                results["source.dH"] / GRAVITATIONAL_CONSTANT * 1.0e5 * results["source.Q"]
            )

            # The pump power should be overestimated compared to the actual head loss due to the
            # fact that we are linearizing a third order equation for hydraulic power instead of
            # the second order for head loss.
            np.testing.assert_array_less(pump_power_post_process, pump_power)

            sum_hp = (
                results["demand.HeatOut.Hydraulic_power"] - results["demand.HeatIn.Hydraulic_power"]
            )
            sum_hp += (
                results["Pipe1.HeatOut.Hydraulic_power"] - results["Pipe1.HeatIn.Hydraulic_power"]
            )
            sum_hp += (
                results["Pipe1_ret.HeatOut.Hydraulic_power"]
                - results["Pipe1_ret.HeatIn.Hydraulic_power"]
            )

            np.testing.assert_allclose(abs(sum_hp), pump_power, atol=1.0e-3)

    def test_heat_network_pipe_split_head_loss(self):
        """
        Heat network: test the piecewise linear weak inequality and equality constraints of the
        head loss approximation.

        Checks:
        - That the head_loss() function does return the expected theoretical dH at a data point
        in the middle of the 1st line segment (dH curve is approximated with 5 linear lines)
        - That the head_loss() function does return a value smaller than a manual linearly
        approximated dH at a data point in the middle of the 1st line segment (dH curve is
        approximated with 5 linear lines)
        - That for the dH value approximated by the code is conservative, in other words greater
        than the theoretical value
        - Compare the optimized dH to the linear calculated value to ensure the specified
        constraint for the head loss linearization is satisfied.
        - For LINEARIZED_N_LINES_EQUALITY:
            - That only one linear line is active for the applicable pipes
            - That the linearized dH value is constraint is satisfied
            - Pipe 4 for has a zero flow rate, but its dH should be the same as pipe 2
        """
        import models.source_pipe_split_sink.src.double_pipe_heat as example
        from models.source_pipe_split_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        # Specify the head loss linearizations to be tested
        for head_loss_option_setting in [
            HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY,
            HeadLossOption.LINEARIZED_N_LINES_EQUALITY,
        ]:
            # Added for case where head loss is modelled via DW
            class SourcePipeSinkDW(SourcePipeSink):
                def energy_system_options(self):
                    options = super().energy_system_options()

                    nonlocal head_loss_option_setting
                    head_loss_option_setting = head_loss_option_setting

                    self.heat_network_settings["head_loss_option"] = head_loss_option_setting

                    self.heat_network_settings["n_linearization_lines"] = 2
                    self.heat_network_settings["minimum_velocity"] = 0.0
                    self.heat_network_settings["minimize_head_losses"] = True

                    return options

            solution = run_esdl_mesido_optimization(
                SourcePipeSinkDW,
                base_folder=base_folder,
                esdl_file_name="sourcesink.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries_import.csv",
            )
            results = solution.extract_results()

            demand_matching_test(solution, results)

            pipes = ["Pipe1", "Pipe2", "Pipe3", "Pipe4"]
            # Only evaluate 1 pipe and 1 timestep for now to reduce the test case computational time
            ipipe = 0
            itime = 0
            v_max = solution.heat_network_settings["maximum_velocity"]
            pipe_diameter = solution.parameters(0)[f"{pipes[ipipe]}.diameter"]
            pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
            temperature = solution.parameters(0)[f"{pipes[ipipe]}.temperature"]
            pipe_length = solution.parameters(0)[f"{pipes[ipipe]}.length"]
            v_points = np.linspace(
                0.0,
                v_max,
                solution.heat_network_settings["n_linearization_lines"] + 1,
            )
            v_inspect = (
                results[f"{pipes[ipipe]}.HeatIn.Q"][itime]
                / solution.parameters(0)[f"{pipes[ipipe]}.area"]
            )

            # Theoretical head loss calc, dH =
            # friction_factor * 8 * pipe_length * volumetric_flow^2 / ( pipe_diameter^5 * g * pi^2)
            dh_theory = (
                darcy_weisbach.friction_factor(
                    v_inspect,
                    pipe_diameter,
                    pipe_wall_roughness,
                    temperature,
                )
                * 8.0
                * pipe_length
                * (v_inspect * np.pi * pipe_diameter**2 / 4.0) ** 2
                / (pipe_diameter**5 * GRAVITATIONAL_CONSTANT * np.pi**2)
            )
            # Approximate dH [m] vs Q [m3/s] with a linear line between between v_points
            # dH_manual_linear = a*Q + b
            # Then use this linear function to calculate the head loss
            delta_dh_theory = darcy_weisbach.head_loss(
                v_points[1], pipe_diameter, pipe_length, pipe_wall_roughness, temperature
            ) - darcy_weisbach.head_loss(
                v_points[0], pipe_diameter, pipe_length, pipe_wall_roughness, temperature
            )

            delta_volumetric_flow = (v_points[1] * np.pi * pipe_diameter**2 / 4.0) - (
                v_points[0] * np.pi * pipe_diameter**2 / 4.0
            )

            a = delta_dh_theory / delta_volumetric_flow
            b = delta_dh_theory - a * delta_volumetric_flow
            dh_manual_linear = a * (v_inspect * np.pi * pipe_diameter**2 / 4.0) + b

            dh_milp_head_loss_function = darcy_weisbach.head_loss(
                v_inspect, pipe_diameter, pipe_length, pipe_wall_roughness, temperature
            )

            np.testing.assert_allclose(dh_theory, dh_milp_head_loss_function)
            np.testing.assert_array_less(dh_milp_head_loss_function, dh_manual_linear)

            if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY:
                np.testing.assert_array_less(
                    dh_manual_linear, -results[f"{pipes[ipipe]}.dH"][itime] + 1e-6
                )
            elif head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                np.testing.assert_allclose(
                    -results[f"{pipes[ipipe]}.dH"][itime], dh_manual_linear, rtol=1e-5, atol=1e-7
                )

            if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                # Check:
                #  - That only one linear line is active for the applicable pipes
                #  - That the linearized dH value is constraint is satisfied
                #  - Pipe 4 for has a zero flow rate, but its dH should be the same as pipe 2
                for pipe in pipes:
                    # Check that only one linear line segment is active for the head loss
                    # linearization
                    if pipe not in ["Pipe4"]:  # Pipe 4 has no flow rate
                        np.testing.assert_allclose(
                            results[f"{pipe}__pipe_linear_line_segment_num_1_neg_discharge"],
                            0.0,
                        )
                        np.testing.assert_allclose(
                            results[f"{pipe}__pipe_linear_line_segment_num_2_neg_discharge"],
                            0.0,
                        )
                        np.testing.assert_allclose(
                            results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"],
                            0.0,
                        )
                        np.testing.assert_allclose(
                            results[f"{pipe}__pipe_linear_line_segment_num_1_pos_discharge"],
                            1.0,
                        )
                        np.testing.assert_allclose(
                            results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"],
                            0.0,
                        )

                    if pipe not in ["Pipe4", "Pipe4_ret"]:
                        velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
                        for ii in range(len(results[f"{pipe}.dH"])):
                            np.testing.assert_array_less(
                                darcy_weisbach.head_loss(
                                    velocities[ii],
                                    pipe_diameter,
                                    pipe_length,
                                    pipe_wall_roughness,
                                    temperature,
                                ),
                                abs(results[f"{pipe}.dH"][ii]),
                            )
                    elif pipe in ["Pipe4", "Pipe4_ret"]:
                        np.testing.assert_allclose(
                            results["Pipe2.dH"][ii], results[f"{pipe}.dH"][ii]
                        )

    def test_gas_network_head_loss(self):
        """
        Gas network: Test the head loss approximation

        Checks:
        - head loss variable vs manually calcuated value
        - that the approximated head loss matches the manually calculated value
        - that linearized dH satisfies the specified constraint
        - that only one linear line segment is active for the head loss linearization
        """

        import models.unit_cases_gas.source_sink.src.run_source_sink as example
        from models.unit_cases_gas.source_sink.src.run_source_sink import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        linear_head_loss_equality = 0.0

        # Specify the head loss linearizations to be tested
        for head_loss_option_setting in [
            HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY,
            HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY,
            HeadLossOption.LINEARIZED_N_LINES_EQUALITY,
        ]:

            class TestSourceSink(GasProblem):
                def energy_system_options(self):
                    options = super().energy_system_options()

                    nonlocal head_loss_option_setting
                    head_loss_option_setting = head_loss_option_setting

                    self.gas_network_settings["head_loss_option"] = head_loss_option_setting
                    if head_loss_option_setting == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY:
                        # This setting is below is not needed for the optmizer, but is used in the
                        # test below.
                        self.gas_network_settings["n_linearization_lines"] = 1  # dot not delete
                        self.gas_network_settings["minimize_head_losses"] = True
                    elif (
                        head_loss_option_setting
                        == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
                    ):
                        self.gas_network_settings["n_linearization_lines"] = 2
                        self.gas_network_settings["minimize_head_losses"] = True
                    if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                        self.gas_network_settings["n_linearization_lines"] = 2
                        self.gas_network_settings["minimize_head_losses"] = True
                        self.gas_network_settings["minimum_velocity"] = 0.0

                    return options

            solution = run_esdl_mesido_optimization(
                TestSourceSink,
                base_folder=base_folder,
                esdl_file_name="source_sink.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries.csv",
            )
            results = solution.extract_results()

            # Check the head loss variable
            np.testing.assert_allclose(
                results["Pipe_4abc.GasOut.H"] - results["Pipe_4abc.GasIn.H"],
                results["Pipe_4abc.dH"],
            )

            pipes = ["Pipe_4abc"]
            v_max = solution.gas_network_settings["maximum_velocity"]
            pipe_diameter = solution.parameters(0)[f"{pipes[0]}.diameter"]
            pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
            # TODO: resolve temperature - >solution.parameters(0)[f"{pipes[0]}.temperature"]
            temperature = 20.0
            pipe_length = solution.parameters(0)[f"{pipes[0]}.length"]
            v_points = [0.0, v_max / solution.gas_network_settings["n_linearization_lines"]]
            v_inspect = results[f"{pipes[0]}.GasOut.Q"] / solution.parameters(0)[f"{pipes[0]}.area"]

            # Approximate dH [m] vs Q [m3/s] with a linear line between between v_points
            # dH_manual_linear = a*Q + b
            # Then use this linear function to calculate the head loss
            delta_dh_theory = darcy_weisbach.head_loss(
                v_points[1],
                pipe_diameter,
                pipe_length,
                pipe_wall_roughness,
                temperature,
                network_type=NetworkSettings.NETWORK_TYPE_GAS,
                pressure=solution.parameters(0)[f"{pipes[0]}.pressure"],
            ) - darcy_weisbach.head_loss(
                v_points[0],
                pipe_diameter,
                pipe_length,
                pipe_wall_roughness,
                temperature,
                network_type=NetworkSettings.NETWORK_TYPE_GAS,
                pressure=solution.parameters(0)[f"{pipes[0]}.pressure"],
            )

            delta_volumetric_flow = (v_points[1] * np.pi * pipe_diameter**2 / 4.0) - (
                v_points[0] * np.pi * pipe_diameter**2 / 4.0
            )

            a = delta_dh_theory / delta_volumetric_flow
            b = delta_dh_theory - a * delta_volumetric_flow
            dh_manual_linear = a * (v_inspect * np.pi * pipe_diameter**2 / 4.0) + b

            # Check that the head loss approximation with 2 linear lines (inequality constraints
            # is < than the linear equality head loss constraint
            if head_loss_option_setting == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY:
                # Check that the aproximated head loss matches the manually calculated value
                np.testing.assert_allclose(dh_manual_linear, -results[f"{pipes[0]}.dH"])
                linear_head_loss_equality = dh_manual_linear
            elif head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY:
                np.testing.assert_array_less(-results[f"{pipes[0]}.dH"], linear_head_loss_equality)
            elif head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                # Check that the approximated head loss matches the manually calculated value
                np.testing.assert_allclose(dh_manual_linear[1], -results[f"{pipes[0]}.dH"][1])

            for pipe in pipes:
                velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
                # linearized dH satisfies the specified constraint
                np.testing.assert_array_less(
                    darcy_weisbach.head_loss(
                        velocities[0], pipe_diameter, pipe_length, pipe_wall_roughness, temperature
                    ),
                    -results[f"{pipe}.dH"][0],
                )
                if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                    # Check that only one linear line segment is active for the head loss
                    # linearization
                    np.testing.assert_allclose(
                        results[f"{pipe}__pipe_linear_line_segment_num_1_neg_discharge"],
                        0.0,
                    )
                    np.testing.assert_allclose(
                        results[f"{pipe}__pipe_linear_line_segment_num_2_neg_discharge"],
                        0.0,
                    )
                    np.testing.assert_allclose(
                        results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"],
                        0.0,
                    )
                    np.testing.assert_allclose(
                        results[f"{pipe}__pipe_linear_line_segment_num_1_pos_discharge"],
                        1.0,
                    )
                    np.testing.assert_allclose(
                        results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"],
                        0.0,
                    )

    def test_gas_network_pipe_split_head_loss(self):
        """
        Gas network: Test the head loss approximation for a parallel pipe network.

        Checks:
        - The head loss variable dH and __head_loss
        - That the gas demand is satisfied
        - Check that the aproximated head loss matches the maunally calculated value
        - That the linearized dH head loss contraint is satisfied
        - That only one linear line segment is active for the head loss linearization. Data points
        exists on both the 1st and 2nd linear line segment

        Notes:
        - The gas demand profile values have been chosen such that 2 data points are present on
        each of the two linearized head loss lines
        """

        import models.unit_cases_gas.source_pipe_split_sink.src.run_source_sink as example
        from models.unit_cases_gas.source_pipe_split_sink.src.run_source_sink import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        # Specify the head loss linearizations to be tested
        for head_loss_option_setting in [
            HeadLossOption.LINEARIZED_N_LINES_EQUALITY,
        ]:

            class TestSourceSink(GasProblem):
                def energy_system_options(self):
                    options = super().energy_system_options()
                    self.gas_network_settings["minimum_velocity"] = 0.0

                    nonlocal head_loss_option_setting
                    head_loss_option_setting = head_loss_option_setting

                    self.gas_network_settings["head_loss_option"] = head_loss_option_setting
                    if head_loss_option_setting == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                        # do not change in value below, see notes above
                        self.gas_network_settings["n_linearization_lines"] = 2
                        self.gas_network_settings["minimize_head_losses"] = True
                    # if statements below are currently not used, potential use in the future
                    elif head_loss_option_setting == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY:
                        self.gas_network_settings["minimize_head_losses"] = True
                    elif (
                        head_loss_option_setting
                        == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
                    ):
                        self.gas_network_settings["n_linearization_lines"] = 2
                        self.gas_network_settings["minimize_head_losses"] = True

                    return options

            solution = run_esdl_mesido_optimization(
                TestSourceSink,
                base_folder=base_folder,
                esdl_file_name="source_sink.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries.csv",
            )
            results = solution.extract_results()

            # Check the head loss variable
            np.testing.assert_allclose(
                results["Pipe1.GasOut.H"] - results["Pipe1.GasIn.H"],
                results["Pipe1.dH"],
            )
            np.testing.assert_allclose(-results["Pipe1.dH"], results["Pipe1.__head_loss"])

            pipes = ["Pipe1"]
            v_max = solution.gas_network_settings["maximum_velocity"]
            pipe_diameter = solution.parameters(0)[f"{pipes[0]}.diameter"]
            pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
            temperature = 20.0
            pipe_length = solution.parameters(0)[f"{pipes[0]}.length"]
            v_points = np.linspace(
                0.0,
                v_max,
                solution.gas_network_settings["n_linearization_lines"] + 1,
            )
            v_inspect = results[f"{pipes[0]}.GasOut.Q"] / solution.parameters(0)[f"{pipes[0]}.area"]

            # Approximate dH [m] vs Q [m3/s] with a linear line between between v_points
            # dH_manual_linear = a*Q + b
            # Then use this linear function to calculate the head loss
            delta_dh_theory = []
            delta_volumetric_flow = []
            a = []
            b = []
            dh_manual_linear = []
            for ii in range(solution.gas_network_settings["n_linearization_lines"]):
                delta_dh_theory.append(
                    darcy_weisbach.head_loss(
                        v_points[ii + 1],
                        pipe_diameter,
                        pipe_length,
                        pipe_wall_roughness,
                        temperature,
                        network_type=NetworkSettings.NETWORK_TYPE_GAS,
                        pressure=solution.parameters(0)[f"{pipes[0]}.pressure"],
                    )
                    - darcy_weisbach.head_loss(
                        v_points[ii],
                        pipe_diameter,
                        pipe_length,
                        pipe_wall_roughness,
                        temperature,
                        network_type=NetworkSettings.NETWORK_TYPE_GAS,
                        pressure=solution.parameters(0)[f"{pipes[0]}.pressure"],
                    )
                )

                delta_volumetric_flow.append(
                    (v_points[ii + 1] * np.pi * pipe_diameter**2 / 4.0)
                    - (v_points[ii] * np.pi * pipe_diameter**2 / 4.0)
                )

                a.append(delta_dh_theory[ii] / delta_volumetric_flow[ii])
                b.append(sum(delta_dh_theory[0:ii]) - a[ii] * sum(delta_volumetric_flow[0:ii]))

            # dh for the 2 data point on the 1st linear line segment
            dh_manual_linear.append(a[0] * (v_inspect[0] * np.pi * pipe_diameter**2 / 4.0) + b[0])
            dh_manual_linear.append(a[0] * (v_inspect[1] * np.pi * pipe_diameter**2 / 4.0) + b[0])
            # dh for the 2 data point on the 2nd linear line segment
            dh_manual_linear.append(a[1] * (v_inspect[2] * np.pi * pipe_diameter**2 / 4.0) + b[1])
            dh_manual_linear.append(a[1] * (v_inspect[3] * np.pi * pipe_diameter**2 / 4.0) + b[1])

            # Gas flow balance
            np.testing.assert_allclose(
                results["GasDemand_a2d8.Gas_demand_mass_flow"],
                solution.get_timeseries("GasDemand_a2d8.target_gas_demand").values,
            )
            # demand_matching_test(solution, results)  # TODO still to be updated for gas networks

            # Check that the aproximated head loss matches the maunally calculated value
            np.testing.assert_allclose(dh_manual_linear, -results["Pipe1.dH"], atol=1e-6)

            for pipe in pipes:
                velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
                np.testing.assert_array_less(
                    darcy_weisbach.head_loss(
                        velocities[0], pipe_diameter, pipe_length, pipe_wall_roughness, temperature
                    ),
                    -results[f"{pipe}.dH"][0],
                )

                # Check that only one linear line segment is active for the head loss linearization
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_1_neg_discharge"],
                    0.0,
                )
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_2_neg_discharge"],
                    0.0,
                )
                # Gas demand for the 1st 2 timesteps fall on the 1st linear line segment
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_1_pos_discharge"][0:2],
                    1.0,
                )
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_1_pos_discharge"][2:4],
                    0.0,
                )
                # Gas demand for the last 2 timesteps fall on the 2nd linear line segment
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"][0:2],
                    0.0,
                )
                np.testing.assert_allclose(
                    results[f"{pipe}__pipe_linear_line_segment_num_2_pos_discharge"][2:4],
                    1.0,
                )

    def test_gas_substation(self):
        """
        Test to check if the gas substation reduces the pressure and the head loss computation
        are correctly performed at the two pressure levels.

        Checks:
        - That the two pipes are at two different pressure levels
        _ That the pipes have the expected head loss given their reference pressures
        """
        import models.multiple_gas_carriers.src.run_multiple_gas_carriers as example
        from models.multiple_gas_carriers.src.run_multiple_gas_carriers import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            GasProblem,
            base_folder=base_folder,
            esdl_file_name="multiple_carriers.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        demand_matching_test(solution, results)

        assert parameters["Pipe1.pressure"] >= parameters["Pipe2.pressure"]

        for pipe in solution.energy_system_components.get("gas_pipe", []):
            dh = results[f"{pipe}.dH"]
            vel = results[f"{pipe}.Q"] / (np.pi * (parameters[f"{pipe}.diameter"] / 2.0) ** 2)
            for i in range(len(solution.times())):
                analytical_dh = (
                    vel[i]
                    / solution.gas_network_settings["maximum_velocity"]
                    * darcy_weisbach.head_loss(
                        solution.gas_network_settings["maximum_velocity"],
                        parameters[f"{pipe}.diameter"],
                        parameters[f"{pipe}.length"],
                        solution.energy_system_options()["wall_roughness"],
                        20.0,
                        network_type=NetworkSettings.NETWORK_TYPE_GAS,
                        pressure=parameters[f"{pipe}.pressure"],
                    )
                )
                np.testing.assert_allclose(abs(dh[i]), abs(analytical_dh), atol=1.0e-6)

    def test_compressor(self):
        """
        Test to check if the gas compressor increases the pressure and the head loss computation
        are correctly performed at the two pressure levels.

        Checks:
        - Demand matching ensuring that there is flow
        - That the two pipes are at two different pressure levels
        _ That the pipes have the expected head loss given their reference pressures
        """
        import models.multiple_gas_carriers.src.run_multiple_gas_carriers as example
        from models.multiple_gas_carriers.src.run_multiple_gas_carriers import GasProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            GasProblem,
            base_folder=base_folder,
            esdl_file_name="compressor.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        demand_matching_test(solution, results)

        assert parameters["Pipe1.pressure"] <= parameters["Pipe2.pressure"]

        for pipe in solution.energy_system_components.get("gas_pipe", []):
            dh = results[f"{pipe}.dH"]
            vel = results[f"{pipe}.Q"] / (np.pi * (parameters[f"{pipe}.diameter"] / 2.0) ** 2)
            for i in range(len(solution.times())):
                analytical_dh = (
                    vel[i]
                    / solution.gas_network_settings["maximum_velocity"]
                    * darcy_weisbach.head_loss(
                        solution.gas_network_settings["maximum_velocity"],
                        parameters[f"{pipe}.diameter"],
                        parameters[f"{pipe}.length"],
                        solution.energy_system_options()["wall_roughness"],
                        20.0,
                        network_type=NetworkSettings.NETWORK_TYPE_GAS,
                        pressure=parameters[f"{pipe}.pressure"],
                    )
                )
                np.testing.assert_allclose(abs(dh[i]), abs(analytical_dh), atol=1.0e-6)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestHeadLoss()
    a.test_heat_network_head_loss()
    a.test_heat_network_pipe_split_head_loss()
    a.test_gas_network_head_loss()
    a.test_gas_network_pipe_split_head_loss()
    a.test_gas_substation()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

import mesido._darcy_weisbach as darcy_weisbach
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.workflows import (
    EndScenarioSizingDiscountedHIGHS,
    EndScenarioSizingHIGHS,
    EndScenarioSizingStagedHIGHS,
    run_end_scenario_sizing,
)
from mesido.workflows.grow_workflow import EndScenarioSizingHeadLossStaged

import numpy as np

from rtctools.util import run_optimization_problem

from utils_tests import demand_matching_test


class TestEndScenarioSizing(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day
        cls.solution = run_optimization_problem(
            EndScenarioSizingHIGHS,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )
        cls.results = cls.solution.extract_results()

    def test_end_scenario_sizing(self):
        """
        Check if the EndScenarioSizingHIGHS sizing workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - Cyclic behaviour for ATES
        - That buffer tank is only used on peak day
        - demand matching
        - Check if TCO goal included the desired cost components.


        Missing:
        - Link ATES t0 utilization to state of charge at end of year for optimizations over one
        year.
        """

        # In the future we want to check the following
        # Is the timeline correctly converted, correct peak day, correct amount of timesteps, etc.
        # Check whether expected assets are disabled
        # Check the optimal size of assets
        # Check the cost breakdown, check whether all the enabled assets are in the cost breakdown
        # Check that computation time is within expected bounds

        # Check whehter the heat demand is matched
        demand_matching_test(self.solution, self.results)

        # Check whether cyclic ates constraint is working
        for a in self.solution.energy_system_components.get("ates", []):
            stored_heat = self.results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)

        # Check whether buffer tank is only active in peak day
        peak_day_indx = self.solution.parameters(0)["peak_day_index"]
        for b in self.solution.energy_system_components.get("heat_buffer", []):
            heat_buffer = self.results[f"{b}.Heat_buffer"]
            for i in range(len(self.solution.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer[i], 0.0, atol=1.0e-6)

        obj = 0.0
        years = self.solution.parameters(0)["number_of_years"]
        for asset in [
            *self.solution.energy_system_components.get("heat_source", []),
            *self.solution.energy_system_components.get("ates", []),
            *self.solution.energy_system_components.get("heat_buffer", []),
            *self.solution.energy_system_components.get("heat_demand", []),
            *self.solution.energy_system_components.get("heat_exchanger", []),
            *self.solution.energy_system_components.get("heat_pump", []),
            *self.solution.energy_system_components.get("heat_pipe", []),
        ]:
            obj += self.results[f"{self.solution._asset_fixed_operational_cost_map[asset]}"] * years
            obj += (
                self.results[f"{self.solution._asset_variable_operational_cost_map[asset]}"] * years
            )
            obj += self.results[f"{self.solution._asset_investment_cost_map[asset]}"]
            obj += self.results[f"{self.solution._asset_installation_cost_map[asset]}"]

        np.testing.assert_allclose(obj / 1.0e6, self.solution.objective_value)

    def test_end_scenario_sizing_staged(self):
        """
        Check if the EndScenarioSizingStagedHIGHS workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - Cyclic behaviour for ATES
        - That buffer tank is only used on peak day
        - demand matching
        - Check if TCO goal included the desired cost components.

        - Compare objective value of staged approach wit non-staged approach

        Compare solution time of 3 scenarios on how to run the optimisation:
        - Staged approach should be solved faster than the unstaged approaches
        - Unstaged approaches, using the general function run_optimization_problem and the
        function run_end_scenario_sizing with staged_pipe_optimization to False should have
        comparable computation times.

        Missing:
        - Link ATES t0 utilization to state of charge at end of year for optimizations over one
        year.
        """
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day

        solution_unstaged = self.solution

        solution_unstaged_2 = run_end_scenario_sizing(
            EndScenarioSizingHIGHS,
            staged_pipe_optimization=False,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        solution_staged = run_end_scenario_sizing(
            EndScenarioSizingStagedHIGHS,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution_staged.extract_results()

        # Check whehter the heat demand is matched
        demand_matching_test(solution_staged, results)

        # Check whether cyclic ates constraint is working
        for a in solution_staged.energy_system_components.get("ates", []):
            stored_heat = results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)

        # Check whether buffer tank is only active in peak day
        peak_day_indx = solution_staged.parameters(0)["peak_day_index"]
        for b in solution_staged.energy_system_components.get("heat_buffer", []):
            heat_buffer = results[f"{b}.Heat_buffer"]
            for i in range(len(solution_staged.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer[i], 0.0, atol=1.0e-6)

        obj = 0.0
        years = solution_staged.parameters(0)["number_of_years"]
        for asset in [
            *solution_staged.energy_system_components.get("heat_source", []),
            *solution_staged.energy_system_components.get("ates", []),
            *solution_staged.energy_system_components.get("heat_buffer", []),
            *solution_staged.energy_system_components.get("heat_demand", []),
            *solution_staged.energy_system_components.get("heat_exchanger", []),
            *solution_staged.energy_system_components.get("heat_pump", []),
            *solution_staged.energy_system_components.get("heat_pipe", []),
        ]:
            obj += results[f"{solution_staged._asset_fixed_operational_cost_map[asset]}"] * years
            obj += results[f"{solution_staged._asset_variable_operational_cost_map[asset]}"] * years
            obj += results[f"{solution_staged._asset_investment_cost_map[asset]}"]
            obj += results[f"{solution_staged._asset_installation_cost_map[asset]}"]

        np.testing.assert_allclose(obj / 1.0e6, solution_staged.objective_value)

        # comparing results of staged and unstaged problem definition. For larger systems there
        # might be a difference in the value but that would either be a difference within the
        # MIPgap (thus checking best bound objective is still smaller or equal than objective of
        # the other problem) or because of some tighter constraints in the staged problem e.g.
        # staged problem slightly higher objective value
        if (
            solution_staged.solver_stats["mip_gap"] == 0.0
            and solution_unstaged.solver_stats["mip_gap"] == 0.0
        ):
            np.testing.assert_allclose(
                solution_staged.objective_value, solution_unstaged.objective_value
            )
        else:
            np.testing.assert_array_less(
                solution_unstaged._priorities_output[1][4]["mip_dual_bound"] - 1e-6,
                solution_staged.objective_value,
            )
            np.testing.assert_array_less(
                solution_staged._priorities_output[3][4]["mip_dual_bound"] - 1e-6,
                solution_unstaged.objective_value,
            )
        # checking time spend on optimisation approaches, the difference between the two unstaged
        # approaches should be smaller than the difference with the staged and unstaged approach.
        # The staged approach should be quickest in solving. The two unstaged approaches should
        # have comparable computation time.
        solution_time_unstaged = sum([i[1] for i in solution_unstaged._priorities_output])
        solution_time_unstaged_2 = sum([i[1] for i in solution_unstaged_2._priorities_output])
        solution_time_staged = sum([i[1] for i in solution_staged._priorities_output])
        np.testing.assert_array_less(
            abs(solution_time_unstaged_2 - solution_time_unstaged),
            abs(solution_time_staged - solution_time_unstaged),
        )
        np.testing.assert_array_less(solution_time_staged, solution_time_unstaged)

    def test_end_scenario_sizing_discounted(self):
        """
        Check if the TestEndScenario sizing workflow is behaving as expected. This is an
        optimization done over a full year with timesteps of 5 days and hour timesteps for the peak
        day.

        Checks:
        - Cyclic behaviour for ATES
        - That buffer tank is only used on peak day
        - demand matching

        Missing:
        - Check if TCO goal included the desired cost components.

        """
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        class TestEndScenarioSizingDiscountedHIGHS(EndScenarioSizingDiscountedHIGHS):
            def solver_options(self):
                options = super().solver_options()
                options["solver"] = "highs"
                highs_options = options["highs"] = {}
                highs_options["mip_rel_gap"] = 0.05
                return options

        # This is an optimization done over a full year with timesteps of 5 days and hour timesteps
        # for the peak day
        solution = run_optimization_problem(
            TestEndScenarioSizingDiscountedHIGHS,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()
        # In the future we want to check the following
        # Is the timeline correctly converted, correct peak day, correct amount of timesteps, etc.
        # Check whether expected assets are disabled
        # Check the optimal size of assets
        # Check the cost breakdown, check whether all the enabled assets are in the cost breakdown
        # Check that computation time is within expected bounds

        # Check whether the heat demand is matched
        demand_matching_test(solution, results)

        # Check whether cyclic ates constraint is working
        for a in solution.energy_system_components.get("ates", []):
            stored_heat = results[f"{a}.Stored_heat"]
            np.testing.assert_allclose(stored_heat[0], stored_heat[-1], atol=1.0)

        # Check whether buffer tank is only active in peak day
        peak_day_indx = solution.parameters(0)["peak_day_index"]
        for b in solution.energy_system_components.get("heat_buffer", []):
            heat_buffer = results[f"{b}.Heat_buffer"]
            for i in range(len(solution.times())):
                if i < peak_day_indx or i > (peak_day_indx + 23):
                    np.testing.assert_allclose(heat_buffer[i], 0.0, atol=1.0e-6)

    def test_end_scenario_sizing_head_loss(self):
        """
        Test is EndScenarioSizingHeadLoss class is behaving as expected. E.g. should behave
        similarly to EndScenarioSizing class but now the linearised inequality Darcy Weisbach
        equations should be included for the head loss.

        Checks:
        - objective (TCO) same order of magnitude
        - head loss is calculated, currently only checked if it is equal or larger than the
        DW head_loss
        """

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        solution = run_end_scenario_sizing(
            EndScenarioSizingHeadLossStaged,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()

        demand_matching_test(solution, results)

        pipes = solution.energy_system_components.get("heat_pipe")
        for pipe in pipes:
            pipe_diameter = solution.parameters(0)[f"{pipes[0]}.diameter"]
            pipe_wall_roughness = solution.energy_system_options()["wall_roughness"]
            temperature = solution.parameters(0)[f"{pipes[0]}.temperature"]
            pipe_length = solution.parameters(0)[f"{pipes[0]}.length"]
            if pipe_diameter > 0.0:
                velocities = results[f"{pipe}.Q"] / solution.parameters(0)[f"{pipe}.area"]
            else:
                velocities = results[f"{pipe}.Q"]  # should be zero
            for ii in range(len(results[f"{pipe}.dH"])):
                if velocities[ii] > 0 and pipe_diameter > 0:
                    np.testing.assert_array_less(
                        darcy_weisbach.head_loss(
                            velocities[ii],
                            pipe_diameter,
                            pipe_length,
                            pipe_wall_roughness,
                            temperature,
                        ),
                        abs(results[f"{pipe}.dH"][ii]),
                    )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestEndScenarioSizing()
    # a.test_end_scenario_sizing()
    a.test_end_scenario_sizing_head_loss()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np


# from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestElectrolyzer(TestCase):
    def test_electrolyzer(self):
        """
        This test is to check the functioning the example with an offshore wind farm in combination
        with an electrolyzer and hydrogen storage.

        Checks:
        - The objective value with the revenue included
        - Check the bounds on the electrolyzer
        - Check the setpoint for the windfarm
        - Check the max production profile of the windfarm
        - Check the electrolyzer inequality constraints formulation

        """
        import models.unit_cases_electricity.electrolyzer.src.example as example
        from models.unit_cases_electricity.electrolyzer.src.example import MILPProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        class MILPProblemSolve(MILPProblem):
            def energy_system_options(self):
                options = super().energy_system_options()
                self.gas_network_settings["pipe_maximum_pressure"] = 100.0  # [bar]
                self.gas_network_settings["pipe_minimum_pressure"] = 0.0
                return options

        solution = run_esdl_mesido_optimization(
            MILPProblem,
            base_folder=base_folder,
            esdl_file_name="h2.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results = solution.extract_results()

        gas_price_profile = "gas.price_profile"
        state = "GasDemand_0cf3.Gas_demand_mass_flow"
        nominal = solution.variable_nominal(state) * np.median(
            solution.get_timeseries(gas_price_profile).values
        )
        gas_revenue = (
            np.sum(
                solution.get_timeseries(gas_price_profile).values
                * results["GasDemand_0cf3.Gas_demand_mass_flow"]
            )
            / nominal
        )

        elec_price_profile = "elec.price_profile"
        state = "ElectricityDemand_9d15.ElectricityIn.Power"
        nominal = solution.variable_nominal(state) * np.median(
            solution.get_timeseries(elec_price_profile).values
        )
        electricity_revenue = (
            np.sum(
                solution.get_timeseries(elec_price_profile).values
                * results["ElectricityDemand_9d15.ElectricityIn.Power"]
            )
            / nominal
        )
        # Check that goal is larger than the revenues as costs are taken into account
        np.testing.assert_array_less(
            -(gas_revenue + electricity_revenue),
            solution.objective_value,
        )
        tol = 1.0e-6
        # Check that the electrolyzer only consumes electricity and does not produce.
        np.testing.assert_array_less(-results["Electrolyzer_fc66.ElectricityIn.Power"], tol)

        # Check that windfarm does not produce more than the specified maximum profile
        ub = solution.get_timeseries("WindPark_7f14.maximum_electricity_source").values
        np.testing.assert_array_less(results["WindPark_7f14.ElectricityOut.Power"], ub + tol)

        # Check that the wind farm setpoint matches with the production
        np.testing.assert_allclose(
            results["WindPark_7f14.ElectricityOut.Power"], ub * results["WindPark_7f14__set_point"]
        )

        # Checks on the storage
        timestep = 3600.0
        rho = solution.parameters(0)["GasStorage_e492.density_max_storage"]
        np.testing.assert_allclose(
            np.diff(results["GasStorage_e492.Stored_gas_mass"]),
            results["GasStorage_e492.Gas_tank_flow"][1:] * rho * timestep,
            rtol=1e-6,
            atol=1e-8,
        )
        np.testing.assert_allclose(results["GasStorage_e492.Stored_gas_mass"][0], 0.0)
        np.testing.assert_allclose(results["GasStorage_e492.Gas_tank_flow"][0], 0.0)

        for cable in solution.energy_system_components.get("electricity_cable", []):
            ub = solution.esdl_assets[solution.esdl_asset_name_to_id_map[f"{cable}"]].attributes[
                "capacity"
            ]
            np.testing.assert_array_less(results[f"{cable}.ElectricityOut.Power"], ub + tol)
            lb = (
                solution.esdl_assets[solution.esdl_asset_name_to_id_map[f"{cable}"]]
                .in_ports[0]
                .carrier.voltage
            )
            tol = 1.0e-2
            np.testing.assert_array_less(lb - tol, results[f"{cable}.ElectricityOut.V"])
            np.testing.assert_array_less(
                results[f"{cable}.ElectricityOut.Power"],
                results[f"{cable}.ElectricityOut.V"] * results[f"{cable}.ElectricityOut.I"] + tol,
            )

        # Electrolyser
        coef_a = solution.parameters(0)["Electrolyzer_fc66.a_eff_coefficient"]
        coef_b = solution.parameters(0)["Electrolyzer_fc66.b_eff_coefficient"]
        coef_c = solution.parameters(0)["Electrolyzer_fc66.c_eff_coefficient"]
        a, b = solution._get_linear_coef_electrolyzer_mass_vs_epower_fit(
            coef_a,
            coef_b,
            coef_c,
            n_lines=3,
            electrical_power_min=0.0,
            electrical_power_max=solution.bounds()["Electrolyzer_fc66.ElectricityIn.Power"][1],
        )
        # TODO: Add test below once the mass flow is coupled to the volumetric flow rate. Currently
        #  the gas network is non-limiting (mass flow not coupled to volumetric flow rate)
        #  np.testing.assert_allclose(results["Electrolyzer_fc66.Gas_mass_flow_out"],
        #                            results["Electrolyzer_fc66.GasOut.Q"] *
        #                            milp_problem.parameters(0)["Electrolyzer_fc66.density"])
        for i in range(len(a)):
            np.testing.assert_array_less(
                results["Electrolyzer_fc66.Gas_mass_flow_out"],
                results["Electrolyzer_fc66.ElectricityIn.Power"] * a[i] + b[i] + 1.0e-3,
            )

        # print(results["Electrolyzer_fc66.ElectricityIn.Power"])
        # print(results["Electrolyzer_fc66.Gas_mass_flow_out"])

        #  -----------------------------------------------------------------------------------------
        # Do cost checks

        # Check variable opex: transport cost 0.1 euro/kg H2
        gas_tranport_cost = sum(
            (
                solution.get_timeseries(elec_price_profile).times[1:]
                - solution.get_timeseries(elec_price_profile).times[0:-1]
            )
            / 3600.0
            * results["Pipe_6ba6.GasOut.mass_flow"][1:]
            * 0.1,
        )
        np.testing.assert_allclose(
            gas_tranport_cost,
            results["GasDemand_0cf3__variable_operational_cost"],
        )

        # Check storage cost fix opex 10 euro/kgH2/year -> 10*23.715 = 237.15euro/m3
        # Storage reserved size = 500m3
        storage_fixed_opex = 237.15 * 500000.0
        np.testing.assert_allclose(
            storage_fixed_opex,
            sum(results["GasStorage_e492__fixed_operational_cost"]),
        )

        # Check electrolyzer fixed opex, based on installed size of 500MW and 10euro/kW
        electrolyzer_fixed_opex = 1.0 * 500.0e6 / 1.0e3
        np.testing.assert_allclose(
            electrolyzer_fixed_opex,
            sum(results["Electrolyzer_fc66__fixed_operational_cost"]),
        )

        # Check electrolyzer investment cost, based on installed size of 500MW and 20euro/kW
        electrolyzer_investment_cost = 20.0 * 500.0e6 / 1.0e3
        np.testing.assert_allclose(
            electrolyzer_investment_cost,
            sum(results["Electrolyzer_fc66__investment_cost"]),
        )
        #  -----------------------------------------------------------------------------------------


if __name__ == "__main__":
    import time

    start_time = time.time()
    test = TestElectrolyzer()
    sol = test.test_electrolyzer()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import fnmatch
import os
import sys
from pathlib import Path
from unittest import TestCase

import esdl
from esdl.esdl_handler import EnergySystemHandler

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.workflows import EndScenarioSizingStagedHIGHS


import numpy as np


class TestUpdatedESDL(TestCase):

    def test_updated_esdl(self):
        """
        Check that the updated ESDL resulting from the optmizer, is correct by using the PoCTutorial
        and the Grow_workflow. This is done for the actual esdl file and the esdl string created by
        MESIDO. Both these resulting optimized energy systems should be identical and it is only
        the MESIDO esdl saving method that differs.

        Checks:
        - That the esdl saving method (direct ESDL file and ESDL string)
        - That the correct number of KPIs have been added
        - That the correct assets have been removed
        - That all the assets have a state=ENABLED
        - The diameter of all the pipes are as expected
        - The aggregation count of the assets, MESIDO problem vs updated ESDL file
        - That the KPI values are represented in the correct units
        - That assets are connected and that the connections per ports were not changed in the
          updated ESDL
        - That the size of the source has been made small. Not checking the exact
        value - not the purpose of these tests
        - The correct number of polygon sub-areas exist
        """

        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.PoCTutorial.src.run_grow_tutorial

        base_folder = (
            Path(examples.PoCTutorial.src.run_grow_tutorial.__file__).resolve().parent.parent
        )
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"

        problem = EndScenarioSizingStagedHIGHS(
            esdl_file_name="PoC Tutorial.esdl",
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
        )
        problem.pre()

        # Load in optimized esdl in the form of esdl string created by MESIDO
        esh = EnergySystemHandler()
        file = os.path.join(base_folder, "model", "PoC Tutorial_GrowOptimized_esdl_string.esdl")
        optimized_energy_system_esdl_string: esdl.EnergySystem = esh.load_file(file)

        # Load in optimized esdl in the form of the actual optimized esdl file created by MESIDO
        esdl_path = os.path.join(base_folder, "model", "PoC Tutorial_GrowOptimized.esdl")
        optimized_energy_system = problem._ESDLMixin__energy_system_handler.load_file(esdl_path)

        optimized_energy_systems = [optimized_energy_system_esdl_string, optimized_energy_system]

        for energy_system in optimized_energy_systems:
            # Test KPIs in optimized ESDL

            # High level checks of KPIs
            number_of_kpis_top_level_in_esdl = 8
            high_level_kpis_euro = [
                "High level cost breakdown [EUR]",
                "Overall cost breakdown [EUR]",
                "CAPEX breakdown [EUR]",
                "OPEX breakdown [EUR]",
                "Area_76a7: Asset cost breakdown [EUR]",
                "Area_9d0f: Asset cost breakdown [EUR]",
                "Area_a58a: Asset cost breakdown [EUR]",
            ]
            high_level_kpis_wh = [
                "Energy production [Wh]",
            ]
            all_high_level_kpis = []
            all_high_level_kpis = high_level_kpis_euro + high_level_kpis_wh

            np.testing.assert_allclose(
                len(energy_system.instance[0].area.KPIs.kpi), number_of_kpis_top_level_in_esdl
            )
            for ii in range(len(energy_system.instance[0].area.KPIs.kpi)):
                kpi_name = energy_system.instance[0].area.KPIs.kpi[ii].name
                np.testing.assert_array_equal(
                    kpi_name in all_high_level_kpis,
                    True,
                    err_msg=f"KPI name {kpi_name} was not expected in the ESDL",
                )
                if kpi_name in high_level_kpis_euro:
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.KPIs.kpi[ii].quantityAndUnit.unit.name,
                        "EURO",
                    )
                elif kpi_name in high_level_kpis_wh:
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.KPIs.kpi[ii].quantityAndUnit.unit.name,
                        "WATTHOUR",
                    )
                else:
                    exit(f"Unexpected KPI name: {kpi_name}")

            # Check the asset quantity
            number_of_assets_in_esdl = 15
            np.testing.assert_allclose(
                len(energy_system.instance[0].area.asset), number_of_assets_in_esdl
            )
            # Check:
            # - that the correct assets were removed
            # - asset state
            # - pipe diameter sizes
            # - asset aggregation count
            # - number of ports
            # - number of connection to a port
            asset_to_be_deleted = ["ResidualHeatSource_76f0", "Pipe_8fa5_ret", "Pipe_8fa5"]
            for ii in range(len(energy_system.instance[0].area.asset)):
                asset_name = energy_system.instance[0].area.asset[ii].name
                # Existance of asset and its state
                np.testing.assert_array_equal(
                    asset_name not in asset_to_be_deleted,
                    True,
                    err_msg=f"Asset name {asset_name} was not expected in the ESDL",
                )
                np.testing.assert_array_equal(
                    energy_system.instance[0].area.asset[ii].state.name == "ENABLED", True
                )

                # Check pipe diameter
                if len(fnmatch.filter([energy_system.instance[0].area.asset[ii].id], "Pipe*")) == 1:
                    if asset_name not in ["Pipe4", "Pipe4_ret", "Pipe5", "Pipe5_ret"]:
                        np.testing.assert_array_equal(
                            energy_system.instance[0].area.asset[ii].diameter.name, "DN400"
                        )
                    else:
                        np.testing.assert_array_equal(
                            energy_system.instance[0].area.asset[ii].diameter.name,
                            "DN300",
                            err_msg=f"Asset name {asset_name} was not expected in the ESDL",
                        )
                    # Check aggregation count
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.asset[ii].aggregationCount,
                        problem.get_aggregation_count_max(asset_name),
                    )
                    # Check the number of ports of the assets are as expected
                    np.testing.assert_array_equal(
                        len(energy_system.instance[0].area.asset[ii].port),
                        len(problem.esdl_assets[asset_name].in_ports)
                        + len(problem.esdl_assets[asset_name].out_ports),
                    )
                    # Check the number of connection to a port
                    energy_system.instance[0].area.asset[ii].port[1].name
                    for iport in range(len(energy_system.instance[0].area.asset[ii].port)):
                        if energy_system.instance[0].area.asset[ii].port[iport].name == "In":
                            np.testing.assert_array_equal(
                                len(
                                    energy_system.instance[0]
                                    .area.asset[ii]
                                    .port[iport]
                                    .connectedTo.items
                                ),
                                len(problem.esdl_assets[asset_name].in_ports),
                            )
                    if asset_name == "ResidualHeatSource_72d7":
                        asset_id = energy_system.instance[0].area.asset[ii].id
                        np.testing.assert_array_less(
                            energy_system.instance[0].area.asset[ii].power,
                            problem.esdl_assets[asset_id].attributes["power"],
                        )

            # High level check on the polygon areas drawn
            number_of_areas_in_esdl = 3
            np.testing.assert_allclose(
                len(energy_system.instance[0].area.area), number_of_areas_in_esdl
            )


if __name__ == "__main__":
    import time

    start_time = time.time()

    a = TestUpdatedESDL()
    a.test_updated_esdl()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestHeat(TestCase):
    def test_heat_loss(self):
        """
        This is a test to check whether the network (pipes) are dissipating milp as we expect.

        Checks:
        - Check that the produced heat is strictly higher than the consumed heat
        - Check for energy conservation in the network

        """
        import models.source_pipe_sink.src.double_pipe_heat as double_pipe_heat
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(double_pipe_heat.__file__).resolve().parent.parent

        case = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = case.extract_results()

        source = results["source.Heat_source"]
        demand = results["demand.Heat_demand"]

        # With non-zero milp losses in pipes, the demand should always be
        # strictly lower than what is produced.
        np.testing.assert_array_less(demand, source)

        demand_matching_test(case, results)
        energy_conservation_test(case, results)
        heat_to_discharge_test(case, results)

    def test_zero_heat_loss(self):
        """
        Check the optimiziation function when the zero heat loss is used.

        Checks:
        - Should check that produced equals consumed.
        - Should check the heat loss variable being zero

        """
        import models.source_pipe_sink.src.double_pipe_heat as double_pipe_heat
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        class Model(SourcePipeSink):
            def energy_system_options(self):
                options = super().energy_system_options()
                options["neglect_pipe_heat_losses"] = True

                return options

        base_folder = Path(double_pipe_heat.__file__).resolve().parent.parent

        case = run_esdl_mesido_optimization(
            Model,
            base_folder=base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )

        results = case.extract_results()
        parameters = case.parameters(0)

        for pipe in case.energy_system_components.get("heat_pipe", []):
            np.testing.assert_allclose(results[f"{pipe}__hn_heat_loss"], 0.0)
            np.testing.assert_allclose(parameters[f"{pipe}.Heat_loss"], 0.0)

        demand_matching_test(case, results)
        energy_conservation_test(case, results)
        heat_to_discharge_test(case, results)


class TestMinMaxPressureOptions(TestCase):
    import models.source_pipe_sink.src.double_pipe_heat as double_pipe_heat
    from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

    base_folder = Path(double_pipe_heat.__file__).resolve().parent.parent
    min_pressure = 4.0
    max_pressure = 12.0
    esdl_file = "sourcesink.esdl"
    input_time_series_file = "timeseries_import.csv"

    class SmallerPipes(SourcePipeSink):
        # We want to force the dynamic pressure in the system to be higher
        # than 12 - 4 = 8 bar (typical upper and lower bounds). We make the
        # pipes smaller in diameter/area to accomplish this. We also adjust
        # the minimum milp delivered by the source, such that a maximum
        # pressure range can force this to go lower than usual.
        def parameters(self, ensemble_member):
            parameters = super().parameters(ensemble_member)
            for p in self.energy_system_components["heat_pipe"]:
                parameters[f"{p}.diameter"] = 0.04
                parameters[f"{p}.area"] = 0.25 * 3.14159265 * parameters[f"{p}.diameter"] ** 2
            return parameters

        def bounds(self):
            bounds = super().bounds()
            bounds["source.Heat_source"] = (0.0, 125_000.0)
            return bounds

        def goals(self):
            return []

    class MinPressure(SmallerPipes):
        def energy_system_options(self):
            options = super().energy_system_options()
            assert "pipe_minimum_pressure" in self.heat_network_settings
            self.heat_network_settings["pipe_minimum_pressure"] = (
                TestMinMaxPressureOptions.min_pressure
            )
            return options

    class MaxPressure(SmallerPipes):
        def energy_system_options(self):
            options = super().energy_system_options()
            assert "pipe_maximum_pressure" in self.heat_network_settings
            options["pipe_maximum_pressure"] = TestMinMaxPressureOptions.max_pressure
            return options

    class MinMaxPressure(SmallerPipes):
        def energy_system_options(self):
            options = super().energy_system_options()
            self.heat_network_settings["pipe_minimum_pressure"] = (
                TestMinMaxPressureOptions.min_pressure
            )
            self.heat_network_settings["pipe_maximum_pressure"] = (
                TestMinMaxPressureOptions.max_pressure
            )
            return options

    def test_min_max_pressure_options(self):
        """
        Check if the min and max pressure options are correctly enforced on the
        optimization. This is achieved by creating a problem where the pressure drop needed
        to match demands is infeasible under the desired pressured range.

        Checks:
        - unbounded problem has requires more pressure drop than allowed by the min and max pressure
        - min pressure
        - max pressure

        """
        case_default = run_esdl_mesido_optimization(
            self.SmallerPipes,
            base_folder=self.base_folder,
            esdl_file_name=self.esdl_file,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file=self.input_time_series_file,
        )
        case_min_pressure = run_esdl_mesido_optimization(
            self.MinPressure,
            base_folder=self.base_folder,
            esdl_file_name=self.esdl_file,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file=self.input_time_series_file,
        )
        case_max_pressure = run_esdl_mesido_optimization(
            self.MaxPressure,
            base_folder=self.base_folder,
            esdl_file_name=self.esdl_file,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file=self.input_time_series_file,
        )
        case_min_max_pressure = run_esdl_mesido_optimization(
            self.MinMaxPressure,
            base_folder=self.base_folder,
            esdl_file_name=self.esdl_file,
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file=self.input_time_series_file,
        )

        def _get_min_max_pressure(case):
            min_head = np.inf
            max_head = -np.inf

            results = case.extract_results()
            for p in case.energy_system_components["heat_pipe"]:
                min_head_in = min(results[f"{p}.HeatIn.H"])
                min_head_out = min(results[f"{p}.HeatOut.H"])
                min_head = min([min_head, min_head_in, min_head_out])

                max_head_in = max(results[f"{p}.HeatIn.H"])
                max_head_out = max(results[f"{p}.HeatOut.H"])
                max_head = max([max_head, max_head_in, max_head_out])

            return min_head / 10.2, max_head / 10.2

        min_, max_ = _get_min_max_pressure(case_default)
        self.assertGreater(max_ - min_, self.max_pressure - self.min_pressure)
        base_objective_value = case_default.objective_value

        min_, max_ = _get_min_max_pressure(case_min_pressure)
        self.assertGreater(min_, self.min_pressure * 0.99)
        self.assertGreater(max_, self.max_pressure)
        self.assertAlmostEqual(case_min_pressure.objective_value, base_objective_value, 4)

        min_, max_ = _get_min_max_pressure(case_max_pressure)
        self.assertLess(min_, self.min_pressure)
        self.assertLess(max_, self.max_pressure * 1.01)
        self.assertAlmostEqual(case_max_pressure.objective_value, base_objective_value, 4)

        min_, max_ = _get_min_max_pressure(case_min_max_pressure)
        self.assertGreater(min_, self.min_pressure * 0.99)
        self.assertLess(max_, self.max_pressure * 1.01)
        target = case_default.get_timeseries("demand.target_heat_demand").values
        self.assertLess(
            np.sum((case_default.extract_results()["demand.Heat_demand"] - target) ** 2),
            np.sum((case_min_max_pressure.extract_results()["demand.Heat_demand"] - target) ** 2),
        )


class TestDisconnectablePipe(TestCase):
    import models.source_pipe_sink.src.double_pipe_heat as double_pipe_heat
    from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

    base_folder = Path(double_pipe_heat.__file__).resolve().parent.parent

    class ModelConnected(SourcePipeSink):
        # We allow the pipe to be disconnectable. We need to be sure that
        # the solution is still feasible (source delivering no heat), so we
        # lower the lower bound.

        def parameters(self, ensemble_member):
            parameters = super().parameters(ensemble_member)
            for p in self.energy_system_components["heat_pipe"]:
                parameters[f"{p}.disconnectable"] = True
            return parameters

        def bounds(self):
            bounds = super().bounds()
            bounds["source.Heat_source"] = (0.0, 125_000.0)
            return bounds

    class ModelDisconnected(ModelConnected):
        def constraints(self, ensemble_member):
            constraints = super().constraints(ensemble_member)

            # Note that we still enforce a _minimum_ velocity in the pipe, if it
            # is connected. So if we force the discharge to zero, that means we
            # force it to be disconnected.
            times = self.times()
            q = self.state_at("Pipe1.Q", times[1], ensemble_member)
            constraints.append((q, 0.0, 0.0))

            return constraints

        def energy_system_options(self):
            options = super().energy_system_options()
            options["heat_loss_disconnected_pipe"] = False
            return options

    class ModelDisconnectedNoHeatLoss(ModelDisconnected):
        def energy_system_options(self):
            options = super().energy_system_options()
            options["heat_loss_disconnected_pipe"] = False
            return options

    def test_disconnected_network_pipe(self):
        """
        Check whether the is_disconnected variable behaves as expected. This variable should be
        True=1 when the flow through the pipe is zero. This is done by running two problems, one
        where the flow is forced to zero and it is expected that the optimization should make the
        pipe disconnected, and one where it is expected for the pipe to stay connected to match
        demand although the pipe is allowed to be disconnected.

        Checks:
        - Sanity check that min velocity is >0
        - Check that pipe stays connected when flow is not forced to zero
        - Check that pipe becomes disconnected when flow is forced to zero

        """
        case_connected = run_esdl_mesido_optimization(
            self.ModelConnected,
            base_folder=self.base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results_connected = case_connected.extract_results()
        q_connected = results_connected["Pipe1.Q"]

        case_disconnected = run_esdl_mesido_optimization(
            self.ModelDisconnected,
            base_folder=self.base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results_disconnected = case_disconnected.extract_results()
        q_disconnected = results_disconnected["Pipe1.Q"]

        # Sanity check, as we rely on the minimum velocity being strictly
        # larger than zero for the discharge constraint to disconnect the
        # pipe.
        self.assertGreater(case_connected.heat_network_settings["minimum_velocity"], 0.0)

        self.assertLess(q_disconnected[1], q_connected[1])
        self.assertAlmostEqual(q_disconnected[1], 0.0, 5)
        np.testing.assert_allclose(results_connected["Pipe1__is_disconnected"], 0.0)
        np.testing.assert_allclose(results_disconnected["Pipe1__is_disconnected"][1], 1.0)

        np.testing.assert_allclose(q_connected[2:], q_disconnected[2:])

    class ModelDisconnectedDarcyWeisbach(ModelDisconnected):
        def energy_system_options(self):
            options = super().energy_system_options()
            self.heat_network_settings["head_loss_option"] = (
                HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
            )
            return options

    def test_disconnected_pipe_darcy_weisbach(self):
        """
        Just a sanity check that the head loss constraints for disconnectable
        pipes works with LINEARIZED_ONE_LINE_EQUALITY as well as LINEARIZED_N_LINES_WEAK_INEQUALITY.

        Checks:
        - That the flow is equal for both types of head loss constraint settings.
        - Check that is_disconnected is set correctly.

        """

        case_linear = run_esdl_mesido_optimization(
            self.ModelDisconnected,
            base_folder=self.base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results_linear = case_linear.extract_results()
        q_linear = results_linear["Pipe1.Q"]

        case_dw = run_esdl_mesido_optimization(
            self.ModelDisconnectedDarcyWeisbach,
            base_folder=self.base_folder,
            esdl_file_name="sourcesink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results_dw = case_dw.extract_results()
        q_dw = results_dw["Pipe1.Q"]

        # Without any constraints on the maximum or minimum head/pressure
        # (loss) in the system, we expect equal results.
        np.testing.assert_allclose(q_linear, q_dw)
        np.testing.assert_allclose(results_dw["Pipe1__is_disconnected"][1], 1.0)



import logging

from rtctools.optimization.timeseries import Timeseries
from rtctools.util import run_optimization_problem

from .. import __version__


def run_heat_network_optimization(heat_class, qht_class, *args, log_level=logging.INFO, **kwargs):
    """
    This function is meant to run the milp milp class and qth non-linear class sequentially. Both
    solve the same network, but the qth class is constrained with the flow direction found in the
    milp optimization. This allows to approximate the full mixed integer non-linear optimization
    with much faster computational times. Note that this does not guarantee optimality, however
    given that the milp milp optimization is reasonably close and conservative it can guarantee
    feasibility.
    """
    logger = logging.getLogger("mesido")
    logger.setLevel(log_level)
    logger.info(f"Using RTC-Tools Heat Network {__version__}.")

    heat_problem = run_optimization_problem(heat_class, *args, log_level=log_level, **kwargs)
    results = heat_problem.extract_results()
    times = heat_problem.times()

    directions = {}

    for p in heat_problem.hot_pipes:
        q_in = results[p + ".HeatIn.Q"]
        q_out = results[p + ".HeatOut.Q"]

        if not heat_problem.parameters(0)[p + ".disconnectable"]:
            # Flow direction is directly related to the sign of the milp
            direction_pipe = (q_in >= 0.0).astype(int) * 2 - 1
        elif heat_problem.parameters(0)[p + ".disconnectable"]:
            direction_pipe = (q_in >= 0.0).astype(int) * 2 - 1
            # Disconnect a pipe when the milp entering the component is only used
            # to account for its milp losses. There are three cases in which this
            # can happen.
            direction_pipe[((q_in > 0.0) & (q_out < 0.0))] = 0
            direction_pipe[((q_in < 0.0) & (q_out > 0.0))] = 0
            direction_pipe[((q_in == 0.0) | (q_out == 0.0))] = 0
        directions[p] = Timeseries(times, direction_pipe)

        # NOTE: The assumption is that the orientation of the cold pipes is such that the flow
        # is always in the same direction as its "hot" pipe companion.
        cold_pipe = heat_problem.hot_to_cold_pipe(p)
        directions[cold_pipe] = directions[p]

    for v in heat_problem.energy_system_components.get("check_valve", []):
        status_valve = (results[f"{v}__status_var"]).round().astype(int)
        directions[v] = Timeseries(times, status_valve)

    for v in heat_problem.energy_system_components.get("control_valve", []):
        directions_valve = (results[f"{v}__flow_direct_var"]).round().astype(int) * 2 - 1
        directions[v] = Timeseries(times, directions_valve)

    buffer_target_discharges = {}
    parameters = heat_problem.parameters(0)

    for b in heat_problem.energy_system_components.get("buffer", []):
        cp = parameters[f"{b}.cp"]
        rho = parameters[f"{b}.rho"]
        heat_flow_rate_to_discharge = 1 / (cp * rho * parameters[f"{b}.dT"])
        buffer_target_discharges[b] = Timeseries(
            times, results[f"{b}.Heat_buffer"] * heat_flow_rate_to_discharge
        )

    qth_problem = run_optimization_problem(
        qht_class,
        *args,
        log_level=log_level,
        flow_directions=directions,
        buffer_target_discharges=buffer_target_discharges,
        **kwargs,
    )

    return heat_problem, qth_problem



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization


class TestEmerge(TestCase):
    def test_emerge_workflow(self):
        """
        This test checks if the emerge workflow is succesfully optimized

        Checks:
        1. check if solver succeeds

        """
        import models.emerge.src.example as example
        from mesido.workflows.emerge import EmergeWorkFlow

        base_folder = Path(example.__file__).resolve().parent.parent

        _ = run_esdl_mesido_optimization(
            EmergeWorkFlow,
            base_folder=base_folder,
            esdl_file_name="emerge.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )



from pathlib import Path
from unittest import TestCase

from mesido._darcy_weisbach import head_loss
from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.network_common import NetworkSettings
from mesido.util import run_esdl_mesido_optimization


import numpy as np

import pandas as pd


class TestHydraulicPower(TestCase):
    def test_hydraulic_power_heat(self):
        """
        Check the workings for the hydraulic power variable.

        Scenario 1. LINEARIZED_N_LINES_WEAK_INEQUALITY (1 line segment)
        Scenario 2. LINEARIZED_ONE_LINE_EQUALITY
        Scenario 3. LINEARIZED_N_LINES_WEAK_INEQUALITY (default line segments = 5)

        Checks:
        - For all scenarios (unless stated otherwise):
            - check that the hydraulic power variable (based on linearized setting) is larger than
            the numerically calculated (post processed)
            - Scenario 1&3: check that the hydraulic power variable = known/verified value for the
            specific case
            - Scenario 1: check that the hydraulic power for the supply and return pipe is the same
            - Scenario 1&2: check that the hydraulic power for these two scenarios are the same
            - Scenario 2: check that the post processed hydraulic power based on flow results
            (voluemtric flow rate * pressure loss) of scenario 1 & 2 are the same.
            - Scenario 3: check that the hydraulic power variable of scenatio 1 > scenario 3, which
            would be expected because scenario 3 has more linear line segments, theerefore the
            approximation would be closer to the theoretical non-linear curve when compared to 1
            linear line approximation of the theoretical non-linear curve.

        Missing:
        - The way the problems are ran and adapted is different compared to the other tests, where
        a global variable is adapted between different runs. I would suggest that we make separate
        problems like we do in the other tests.
        - Also I would prefer using the results directly in this test instead of calling the
        df_MILP.
        - See if the hard coded values can be avoided.

        """
        import models.pipe_test.src.run_hydraulic_power as run_hydraulic_power
        from models.pipe_test.src.run_hydraulic_power import (
            HeatProblem,
        )

        # Settings
        base_folder = Path(run_hydraulic_power.__file__).resolve().parent.parent
        run_hydraulic_power.comp_vars_vals = {
            "pipe_length": [25000.0],  # [m]
        }
        run_hydraulic_power.comp_vars_init = {
            "pipe_length": 0.0,  # [m]
            "heat_demand": [3.95 * 10**6, 3.95 * 10**6],  # [W]
            "pipe_DN_MILP": 300,  # [mm]
        }
        standard_columns_specified = [
            "Pipe1_supply_dPress",
            "Pipe1_return_dPress",
            "Pipe1_supply_Q",
            "Pipe1_return_Q",
            "Pipe1_supply_mass_flow",
            "Pipe1_return_mass_flow",
            "Pipe1_supply_flow_vel",
            "Pipe1_return_flow_vel",
            "Pipe1_supply_dT",
            "Pipe1_return_dT",
            "Heat_source",
            "Heat_demand",
            "Heat_loss",
            "pipe_length",
        ]

        # Initialize variables
        run_hydraulic_power.ThermalDemand = run_hydraulic_power.comp_vars_init["heat_demand"]
        run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_init[
            "pipe_length"
        ]
        run_hydraulic_power.manual_set_pipe_DN_diam_MILP = run_hydraulic_power.comp_vars_init[
            "pipe_DN_MILP"
        ]
        # ----------------------------------------------------------------------------------------
        # 3 MILP simulations with the only difference being the linear head loss setting:
        # - LINEARIZED_N_LINES_WEAK_INEQUALITY (1 line segment)
        # - LINEARIZED_ONE_LINE_EQUALITY
        # - LINEARIZED_N_LINES_WEAK_INEQUALITY (default line segments = 5)
        # ----------------------------------------------------------------------------------------
        # Run MILP with LINEARIZED_N_LINES_WEAK_INEQUALITY head loss setting and 1 line segement
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
        run_hydraulic_power.n_linearization_lines_setting = 1

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_esdl_mesido_optimization(
                HeatProblem,
                base_folder=base_folder,
                esdl_file_name="test_simple.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries_import.xml",
            )

        hydraulic_power_post_process_dw_1 = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][0] * abs(
            run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0]
        )
        hydraulic_power_dw_1 = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (hydraulic_power_dw_1) of the 3rd order function should
        # overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_dw_1,
            hydraulic_power_dw_1,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Compare hydraulic power, for an one hour timeseries with a specific demand, to a hard
        # coded value which originates from runnning MILP without big_m method being implemented,
        # during the comparison of MILP and a high-fidelity code
        # FIXME: this value from high-fidelity code needs to be checked, due to changes in the setup
        # of the heat_to_discharge constraints, the volumetric flow has increased, resulting in
        # larger pressure drops.
        np.testing.assert_allclose(128001.23151838078, hydraulic_power_dw_1, atol=10)
        np.testing.assert_allclose(
            run_hydraulic_power.df_MILP["Pipe1_return_Hydraulic_power"][0],
            run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0],
            rtol=1e-2,
        )
        # ----------------------------------------------------------------------------------------
        # Rerun MILP with LINEARIZED_ONE_LINE_EQUALITY head loss setting
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)  # empty df
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_esdl_mesido_optimization(
                HeatProblem,
                base_folder=base_folder,
                esdl_file_name="test_simple.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries_import.xml",
            )

        hydraulic_power_post_process_linear = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][
            0
        ] * abs(run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0])
        hydraulic_power_linear = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (1 line segment) of the 3rd order function should
        # overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_linear,
            hydraulic_power_linear,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Hydraulic hydraulic =  delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The predicted hydraulic power should be the same when the delta pressure is approximated
        # by a linear segment, via 2 different head loss setting options. Head loss setting =
        # HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY and
        # HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY (with 1 linear segment)
        np.testing.assert_allclose(
            hydraulic_power_post_process_linear,
            hydraulic_power_post_process_dw_1,
            rtol=1e-7,
            err_msg="Values should be the same",
        )
        # Hydraulic hydraulic =  delta pressure * Q = f(Q^3)
        # The predicted hydraulic power should be the same if it is approximated by a linear segment
        # , via 2 different head loss setting options. Head loss setting =
        # HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY and
        # HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY (with 1 linear segment)
        np.testing.assert_allclose(
            hydraulic_power_linear, hydraulic_power_dw_1, err_msg="Values should be the same"
        )
        # ----------------------------------------------------------------------------------------
        # Rerun MILP with DW head loss setting, and default line segments
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)  # empty df
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
        run_hydraulic_power.n_linearization_lines_setting = 5

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_esdl_mesido_optimization(
                HeatProblem,
                base_folder=base_folder,
                esdl_file_name="test_simple.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="timeseries_import.xml",
            )

        hydraulic_power_post_process_dw = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][0] * abs(
            run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0]
        )
        hydraulic_power_dw = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (default number of line segments) of the 3rd order function
        # should overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation (default number of line segments) of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_dw,
            hydraulic_power_dw,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The approximation of the 3rd order function via 5 line segments (dafault value) should be
        # better compared to 1 line segment approximation thereof. The latter will result in an
        # overstimated prediction
        np.testing.assert_array_less(
            hydraulic_power_dw,
            hydraulic_power_dw_1,
            "5 line segments predicted hydraulic power > hydraulic_power with 1 line segment",
        )
        # Compare hydraulic power, for an one hour timeseries with a specific demand, to a hard
        # coded value which originates from runnning MILP without big_m method being implemented,
        # during the comparison of MILP and a high-fidelity code
        # FIXME: this value from high-fidelity code needs to be checked, due to changes in the setup
        #  of the heat_to_discharge constraints, the volumetric flow has increased, resulting in
        #  larger pressure drops.
        np.testing.assert_allclose(
            5332.57631593844,
            hydraulic_power_dw,
            atol=10.0,
        )

    def test_hydraulic_power_gas(self):
        """
        Checks the logic for the hydraulic power of gas pipes.
        - checks if value on linearized lines, for multiple lines
        - checks if hydraulic power is 0 at end of the pipe
        - checks if differences of in/out port is equal to the added hydraulic power of that pipe
        - checks absolutae value of the hydraulic power loss over a line
        """
        import models.unit_cases_gas.source_sink.src.run_source_sink as run_source_sink
        from models.unit_cases_gas.source_sink.src.run_source_sink import (
            GasProblem,
        )

        # Settings
        base_folder = Path(run_source_sink.__file__).resolve().parent.parent

        class GasProblemHydraulic(GasProblem):
            def read(self):
                super().read()

                for d in self.energy_system_components["gas_demand"]:
                    new_timeseries = self.get_timeseries(f"{d}.target_gas_demand").values * 4e4
                    self.set_timeseries(f"{d}.target_gas_demand", new_timeseries)

            def energy_system_options(self):
                options = super().energy_system_options()

                self.gas_network_settings["head_loss_option"] = (
                    HeadLossOption.LINEARIZED_N_LINES_EQUALITY
                )
                self.gas_network_settings["n_linearization_lines"] = 3
                self.gas_network_settings["minimum_velocity"] = 0.0
                self.gas_network_settings["minimize_head_losses"] = True

                return options

        solution = run_esdl_mesido_optimization(
            GasProblemHydraulic,
            base_folder=base_folder,
            esdl_file_name="source_sink.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        # TODO: add check on values for hydraulic power.
        results = solution.extract_results()
        parameters = solution.parameters(0)

        pipe = "Pipe_4abc"
        pipe_hp_in = results[f"{pipe}.GasIn.Hydraulic_power"]
        pipe_hp_out = results[f"{pipe}.GasOut.Hydraulic_power"]
        pipe_hp = results[f"{pipe}.Hydraulic_power"]

        pipe_mass = results["Pipe_4abc.GasIn.mass_flow"]

        # due to non linearity, every timestep on new linearized line, a doubled mass flow should
        # result in more than doubled hydraulic power
        np.testing.assert_array_less(
            (pipe_hp[0] + 1e-6) * (pipe_mass[1] / pipe_mass[0]), pipe_hp[1]
        )
        np.testing.assert_array_less(
            (pipe_hp[1] + 1e-6) * (pipe_mass[2] / pipe_mass[1]), pipe_hp[2]
        )

        np.testing.assert_allclose(pipe_hp, pipe_hp_in - pipe_hp_out)
        np.testing.assert_allclose(0, pipe_hp_out)

        # TODO: use mass flow to get calculated hydraulic power

        rho = parameters[f"{pipe}.rho"]
        d = parameters[f"{pipe}.diameter"]
        area = parameters[f"{pipe}.area"]
        length = parameters[f"{pipe}.length"]
        pressure = parameters[f"{pipe}.pressure"]
        wall_roughness = solution.energy_system_options()["wall_roughness"]
        v_max = solution.gas_network_settings["maximum_velocity"]
        temperature = 20.0

        v_inspect = results[f"{pipe}.GasOut.Q"] / solution.parameters(0)[f"{pipe}.area"]

        calc_hp_accurate = [
            rho
            * GRAVITATIONAL_CONSTANT
            * v
            * area
            * head_loss(
                v,
                d,
                length,
                wall_roughness,
                temperature,
                network_type=NetworkSettings.NETWORK_TYPE_GAS,
                pressure=pressure,
            )
            for v in v_inspect
        ]
        np.testing.assert_array_less(calc_hp_accurate, pipe_hp)

        v_points = [
            i * v_max / solution.gas_network_settings["n_linearization_lines"]
            for i in range(solution.gas_network_settings["n_linearization_lines"] + 1)
        ]
        calc_hp_v_points = [
            rho
            * GRAVITATIONAL_CONSTANT
            * v
            * area
            * head_loss(
                v,
                d,
                length,
                wall_roughness,
                temperature,
                network_type=NetworkSettings.NETWORK_TYPE_GAS,
                pressure=pressure,
            )
            for v in v_points
        ]
        v_points_volumetric = np.asarray(v_points) * np.pi * d**2 / 4.0
        a = np.diff(calc_hp_v_points) / np.diff(v_points_volumetric)
        b = calc_hp_v_points[1:] - a * v_points_volumetric[1:]

        np.testing.assert_allclose(pipe_hp[0], a[0] * results[f"{pipe}.GasOut.Q"][0] + b[0])
        np.testing.assert_allclose(pipe_hp[1], a[1] * results[f"{pipe}.GasOut.Q"][1] + b[1])
        np.testing.assert_allclose(pipe_hp[2], a[2] * results[f"{pipe}.GasOut.Q"][2] + b[2])

    def test_hydraulic_power_gas_multi_demand(self):
        """
        Checks the logic for the hydraulic power of gas pipes.
        - checks if value on linearized lines, for multiple lines
        - checks if hydraulic power is 0 at end of the pipe if connected to a demand
        - checks if hydraulic power at start of line connected to a producers is larger than 0
        - checks if differences of in/out port is equal to the added hydraulic power of that pipe
        """
        import models.unit_cases_gas.multi_demand_source_node.src.run_test as run_test
        from models.unit_cases_gas.multi_demand_source_node.src.run_test import (
            GasProblem,
        )

        # Settings
        base_folder = Path(run_test.__file__).resolve().parent.parent

        class GasProblemHydraulic(GasProblem):
            def read(self):
                super().read()

                for d in self.energy_system_components["gas_demand"]:
                    new_timeseries = self.get_timeseries(f"{d}.target_gas_demand").values * 1.5
                    self.set_timeseries(f"{d}.target_gas_demand", new_timeseries)

            def energy_system_options(self):
                options = super().energy_system_options()

                self.gas_network_settings["head_loss_option"] = (
                    HeadLossOption.LINEARIZED_N_LINES_EQUALITY
                )
                self.gas_network_settings["n_linearization_lines"] = 3
                self.gas_network_settings["minimum_velocity"] = 0.0
                self.gas_network_settings["minimize_head_losses"] = True

                return options

        solution = run_esdl_mesido_optimization(
            GasProblemHydraulic,
            base_folder=base_folder,
            esdl_file_name="test.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        # TODO: add check on values for hydraulic power.
        results = solution.extract_results()

        pipes = ["Pipe_7c53", "Pipe_f1a4", "Pipe_0e39", "Pipe_c50f"]
        for pipe in pipes:
            pipe_hp_in = results[f"{pipe}.GasIn.Hydraulic_power"]
            pipe_hp_out = results[f"{pipe}.GasOut.Hydraulic_power"]
            pipe_hp = results[f"{pipe}.Hydraulic_power"]

            pipe_mass = results[f"{pipe}.GasIn.mass_flow"]

            np.testing.assert_allclose(pipe_hp, pipe_hp_in - pipe_hp_out)
            if pipe in ["Pipe_7c53", "Pipe_c50f"]:
                # connected to demand, thus hydraulic_power should be 0
                np.testing.assert_allclose(0, pipe_hp_out)

            v_max = solution.gas_network_settings["maximum_velocity"]

            v_inspect = results[f"{pipe}.GasOut.Q"] / solution.parameters(0)[f"{pipe}.area"]
            v_points = [
                i * v_max / solution.gas_network_settings["n_linearization_lines"]
                for i in range(solution.gas_network_settings["n_linearization_lines"] + 1)
            ]

            # due to non linearity, every timestep on new linearized line, a doubled mass flow
            # should result in more than doubled hydraulic power
            v_inspect_line_ind = []
            for k in range(len(v_inspect)):
                for i in range(len(v_points)):
                    if v_points[i] + 1e-6 < v_inspect[k] <= v_points[i + 1] + 1e-6:
                        v_inspect_line_ind.append(i)
            ind_check = 0
            for k in range(len(v_inspect) - 1):
                if v_inspect_line_ind[k] == v_inspect_line_ind[k + 1]:
                    np.testing.assert_allclose(
                        pipe_hp[k] * pipe_mass[k + 1] / pipe_mass[k], pipe_hp[k + 1]
                    )
                elif v_inspect_line_ind[k] < v_inspect_line_ind[k + 1]:
                    np.testing.assert_array_less(
                        (pipe_hp[k] + 1e-6) * pipe_mass[k + 1] / pipe_mass[k], pipe_hp[k + 1]
                    )
                    ind_check += 1
                else:
                    raise RuntimeError(
                        "For this test to succeed the flow should increase over time"
                    )
            np.testing.assert_array_less(
                0.5, ind_check, f"{pipe} is not checked between multiple lines"
            )

        # balance of hydraulic power
        pipes_demand = ["Pipe_7c53", "Pipe_c50f"]
        pipes_source = ["Pipe_0e39", "Pipe_f1a4"]
        balance = np.zeros(3)
        for p in pipes_demand:
            balance += results[f"{p}.GasIn.Hydraulic_power"]
        for p in pipes_source:
            balance -= results[f"{p}.GasOut.Hydraulic_power"]
        np.testing.assert_allclose(balance, 0.0, atol=1e-6)

        for node, connected_pipes in solution.energy_system_topology.gas_nodes.items():
            hydraulic_sum = 0.0

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                hydraulic_sum += results[f"{node}.GasConn[{i_conn+1}].Q"] * orientation

            np.testing.assert_allclose(hydraulic_sum, 0.0, atol=1.0e-3)

        for d in solution.energy_system_components.get("gas_demand"):
            np.testing.assert_allclose(results[f"{d}.GasIn.Hydraulic_power"], 0.0)
        for d in solution.energy_system_components.get("gas_source"):
            np.testing.assert_array_less(0.0, results[f"{d}.GasOut.Hydraulic_power"])


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestHydraulicPower()
    a.test_hydraulic_power()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import math

from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal.heat_component import BaseAsset

from numpy import nan

from .heat_two_port import HeatTwoPort


class HeatBuffer(HeatTwoPort, BaseAsset):
    """
    The buffer component is to model milp storage in a tank. This means that we model a tank of hot
    water being filled and radiating milp away (heat loss) over the hot surfaces. We assume that the
    hot surfaces are those in contact with hot water.

    Like all storage assets we enforce that they must be connected as a demand. The heat to
    discharge constraints are set in the HeatPhysicsMixin, where we use a big_m formulation to
    enforce the correct constraints depending on whether the buffer is charging or discharging.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "heat_buffer"

        self.Q_nominal = 1.0
        self.T_supply = nan
        self.T_return = nan
        self.T_supply_id = -1
        self.T_return_id = -1
        self.dT = self.T_supply - self.T_return
        self.cp = 4200.0
        self.rho = 988.0
        self.Heat_nominal = self.cp * self.rho * self.dT * self.Q_nominal
        self.nominal_pressure = 16.0e5
        self.minimum_pressure_drop = 1.0e5  # 1 bar of pressure drop
        self.pump_efficiency = 0.5

        self.heat_transfer_coeff = 1.0
        self.height = 5.0
        self.radius = 10.0
        self.volume = math.pi * self.radius**2 * self.height
        self.heat_loss_coeff = 2 * self.heat_transfer_coeff / (self.radius * self.rho * self.cp)
        # The hot/cold tank can have a lower bound on its volume.
        # Meaning that they might always be, for e.g., 5% full.
        self.min_fraction_tank_volume = 0.05

        # Initial values
        self.init_V_hot_tank = nan
        self.init_Heat = nan

        # Minimum/maximum values
        self.min_stored_heat = (
            self.volume * self.min_fraction_tank_volume * self.dT * self.cp * self.rho
        )
        self.max_stored_heat = (
            self.volume * (1 - self.min_fraction_tank_volume) * self.dT * self.cp * self.rho
        )

        # Stored_heat is the heat that is contained in the buffer.
        # Heat_buffer is the amount of heat added to or extracted from the buffer
        # per timestep.
        # HeatHot (resp. HeatCold) is the amount of heat added or extracted from
        # the hot (resp. cold) line.
        # As by construction the cold line should have zero heat, we fix HeatCold to zero.
        # Thus Heat_buffer = HeatHot = der(Stored_heat).
        # We connect a buffer as an demand, meaning that flow and Heat_buffer are positive under
        # charging and negative under discharge
        self.add_variable(Variable, "Heat_buffer", nominal=self.Heat_nominal)
        # Assume the storage fills in about an hour at typical rate
        self._typical_fill_time = 3600.0
        self._nominal_stored_heat = self.Heat_nominal * self._typical_fill_time
        self.add_variable(
            Variable,
            "Stored_heat",
            min=self.min_stored_heat,
            max=self.max_stored_heat,
            nominal=self._nominal_stored_heat,
        )
        self.add_variable(Variable, "Q", nominal=self.Q_nominal)
        # For nicer constraint coefficient scaling, we shift a bit more error into
        # the state vector entry of `Heat_loss`. In other words, with a factor of
        # 10.0, we aim for a state vector entry of ~0.1 (instead of 1.0)
        self._heat_loss_error_to_state_factor = 10.0
        self._nominal_heat_loss = (
            self._nominal_stored_heat * self.heat_loss_coeff * self._heat_loss_error_to_state_factor
        )
        self.add_variable(Variable, "Heat_loss", min=0.0, nominal=self._nominal_heat_loss)
        self.add_variable(Variable, "Heat_flow", nominal=self.Heat_nominal)
        self.add_variable(
            Variable, "Pump_power", min=0.0, nominal=self.Q_nominal * self.nominal_pressure
        )

        self._heat_loss_eq_nominal_buf = (self.Heat_nominal * self._nominal_heat_loss) ** 0.5

        self.add_variable(Variable, "dH")
        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation(self.HeatIn.Q - self.HeatOut.Q)
        self.add_equation(self.Q - self.HeatOut.Q)

        # Heat stored in the buffer
        self.add_equation(
            (self.der(self.Stored_heat) - self.Heat_buffer + self.Heat_loss)
            / self._heat_loss_eq_nominal_buf
        )
        self.add_equation(
            (self.Heat_loss - self.Stored_heat * self.heat_loss_coeff) / self._nominal_heat_loss
        )
        self.add_equation(
            (self.Heat_buffer - (self.HeatIn.Heat - self.HeatOut.Heat)) / self.Heat_nominal
        )
        self.add_equation((self.Heat_flow - self.Heat_buffer) / self.Heat_nominal)



import datetime
import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Set

import esdl
from esdl.profiles.influxdbprofilemanager import ConnectionSettings
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager
from esdl.units.conversion import ENERGY_IN_J, POWER_IN_W, convert_to_unit

from mesido.esdl.common import Asset

import numpy as np

import pandas as pd

import rtctools.data.pi
from rtctools.data.storage import DataStore


logger = logging.getLogger()

influx_cred_map = {"wu-profiles.esdl-beta.hesi.energy:443": ("warmingup", "warmingup")}


class _ProfileParserException(Exception):
    pass


class BaseProfileReader:
    component_type_to_var_name_map: dict = {
        "cold_demand": ".target_cold_demand",
        "heat_demand": ".target_heat_demand",
        "heat_source": ".maximum_heat_source",
        "electricity_demand": ".target_electricity_demand",
        "electricity_source": ".maximum_electricity_source",
        "gas_demand": ".target_gas_demand",
        "gas_source": ".maximum_gas_source",
    }
    carrier_profile_var_name: str = ".price_profile"

    def __init__(self, energy_system: esdl.EnergySystem, file_path: Optional[Path]):
        self._profiles: Dict[int, Dict[str, np.ndarray]] = defaultdict(dict)
        self._energy_system: esdl.EnergySystem = energy_system
        self._file_path: Optional[Path] = file_path
        self._reference_datetimes: Optional[pd.DatetimeIndex] = None

    def read_profiles(
        self,
        io: DataStore,
        energy_system_components: Dict[str, Set[str]],
        esdl_asset_id_to_name_map: Dict[str, str],
        esdl_assets: Dict[str, Asset],
        carrier_properties: Dict[str, Dict],
        ensemble_size: int,
    ) -> None:
        """
        This function takes a datastore and a dictionary of milp network components and loads a
        profile for each demand and source in the provided milp network components into the
        datastore. If no profile is available the following happens:
        - for sources, no target profile is set
        - for demands a default profile is loaded equal to the power of the demand asset
        Note that at least one profile must be provided to determine the start and end times of the
        optimization horizon.

        Parameters
        ----------
        io : Datastore in which the profiles will be saved
        energy_system_components :   Dictionary of the components of the network, should
                                    contain at least every component for which a profile
                                    needs to be loaded
        esdl_asset_id_to_name_map : Dictionary that maps asset ids to asset names,
                                    this is required when reading from an XML
        esdl_assets : Dictionary mapping asset IDs to loaded ESDL assets
        esdl_carriers: Dictionary mapping carrier IDs to its properties
        ensemble_size :     Integer denoting the size of the set of scenarios to
                            optimize. Currently only XML inputs support loading a
                            different profile for different ensemble members

        Returns
        -------
        None
        """
        self._load_profiles_from_source(
            energy_system_components=energy_system_components,
            esdl_asset_id_to_name_map=esdl_asset_id_to_name_map,
            carrier_properties=carrier_properties,
            ensemble_size=ensemble_size,
        )

        try:
            io.reference_datetime = self._reference_datetimes[0]
        except AttributeError:
            raise RuntimeError(
                "No profiles were provided so no timeframe for the profiles could be deduced"
            )

        esdl_asset_names_to_ids = dict(
            zip(esdl_asset_id_to_name_map.values(), esdl_asset_id_to_name_map.keys())
        )

        for ensemble_member in range(ensemble_size):
            for component_type, var_name in self.component_type_to_var_name_map.items():
                for component in energy_system_components.get(component_type, []):
                    profile = self._profiles[ensemble_member].get(component + var_name, None)
                    if profile is not None:
                        values = profile
                    else:
                        if "heat_demand" not in component_type:
                            # We don't set a default profile for source targets
                            continue
                        logger.warning(
                            f"No profile provided for {component=} and "
                            f"{ensemble_member=}, using the assets power value instead"
                        )
                        asset_power = esdl_assets[esdl_asset_names_to_ids[component]].attributes[
                            "power"
                        ]
                        values = np.array([asset_power] * len(self._reference_datetimes))

                    io.set_timeseries(
                        variable=component + var_name,
                        datetimes=self._reference_datetimes,
                        values=values,
                        ensemble_member=ensemble_member,
                    )
            for properties in carrier_properties.values():
                carrier_name = properties["name"]
                profile = self._profiles[ensemble_member].get(
                    carrier_name + self.carrier_profile_var_name, None
                )
                if profile is not None:
                    logger.debug(
                        f"Setting price profile for carrier named {carrier_name} " f"to {profile}"
                    )
                    io.set_timeseries(
                        variable=carrier_name + self.carrier_profile_var_name,
                        datetimes=self._reference_datetimes,
                        values=profile,
                        ensemble_member=ensemble_member,
                    )

    def _load_profiles_from_source(
        self,
        energy_system_components: Dict[str, Set[str]],
        esdl_asset_id_to_name_map: Dict[str, str],
        carrier_properties: Dict[str, Dict],
        ensemble_size: int,
    ) -> None:
        """
        This function must be implemented by the child. It must load the available
        profiles for demands and sources from the correct source and saves them in the _profiles
        attribute. It must also set the _reference_datetime_index attribute to the correct
        index to be used in the DataStore when loading the profiles

        Parameters
        ----------
        energy_system_components :   Dictionary of the components of the network, should
                                    contain at least every component for which a profile
                                    needs to be loaded
        esdl_asset_id_to_name_map : Dictionary that maps asset ids to asset names,
                                    this is required when reading from an XML
        ensemble_size :     Integer denoting the size of the set of scenarios to
                            optimize. Currently only XML inputs support loading a
                            different profile for different ensemble members

        Returns
        -------
        None
        """
        raise NotImplementedError


class InfluxDBProfileReader(BaseProfileReader):
    asset_type_to_variable_name_conversion = {
        esdl.esdl.HeatingDemand: ".target_heat_demand",
        esdl.esdl.HeatProducer: ".maximum_heat_source",
        esdl.esdl.ElectricityDemand: ".target_electricity_demand",
        esdl.esdl.ElectricityProducer: ".maximum_electricity_source",
        esdl.esdl.GasDemand: ".target_gas_demand",
        esdl.esdl.GasProducer: ".maximum_gas_source",
    }

    def __init__(self, energy_system: esdl.EnergySystem, file_path: Optional[Path]):
        super().__init__(energy_system=energy_system, file_path=file_path)
        self._df = pd.DataFrame()

    def _load_profiles_from_source(
        self,
        energy_system_components: Dict[str, Set[str]],
        esdl_asset_id_to_name_map: Dict[str, str],
        carrier_properties: Dict[str, Dict],
        ensemble_size: int,
    ) -> None:
        profiles: Dict[str, np.ndarray] = dict()
        logger.info("Reading profiles from InfluxDB")
        self._reference_datetimes = None
        for profile in [
            x for x in self._energy_system.eAllContents() if isinstance(x, esdl.InfluxDBProfile)
        ]:
            series = self._load_profile_timeseries_from_database(profile=profile)
            self._check_profile_time_series(profile_time_series=series, profile=profile)
            if self._reference_datetimes is None:
                # TODO: since the previous function ensures it's a date time index, I'm not sure
                #  how to get rid of this type checking warning
                self._reference_datetimes = series.index
            else:
                if not all(series.index == self._reference_datetimes):
                    raise RuntimeError(
                        f"Obtained a profile for asset {profile.field} with a "
                        f"timeseries index that doesn't match the timeseries of "
                        f"other assets. Please ensure that the profile that is "
                        f"specified to be loaded for each asset covers exactly the "
                        f"same timeseries. "
                    )
            converted_dataframe = self._convert_profile_to_correct_unit(
                profile_time_series=series, profile=profile
            )

            container = profile.eContainer()
            if isinstance(container, esdl.Commodity):
                variable_suffix = self.carrier_profile_var_name
                var_base_name = container.name
            elif isinstance(container, esdl.Port):
                asset = container.energyasset
                var_base_name = asset.name
                try:
                    variable_suffix = self.asset_type_to_variable_name_conversion[type(asset)]
                except KeyError:
                    raise RuntimeError(
                        f"The asset {profile.field} is of type {type(asset)} which is "
                        f"currently not supported to have a profile to be loaded "
                        f"from the database."
                    )
            else:
                raise RuntimeError(
                    f"Got a profile for a {container}. Currently only profiles "
                    f"for assets and commodities are supported"
                )
            profiles[var_base_name + variable_suffix] = converted_dataframe * profile.multiplier

        for idx in range(ensemble_size):
            self._profiles[idx] = profiles.copy()

    # @staticmethod
    def _load_profile_timeseries_from_database(self, profile: esdl.InfluxDBProfile) -> pd.Series:
        """
        Function to load the profiles from an InfluxDB. Returns a timeseries with the data for
        the asset.

        Parameters
        ----------
        profile : Input InfluxDBProfile for the asset in the ESDL for which a profile should be read

        Returns
        -------
        A pandas Series of the profile for the asset.
        """
        if profile.id in self._df:
            return self._df[profile.id]

        profile_host = profile.host

        ssl_setting = False
        if "https" in profile_host:
            profile_host = profile_host[8:]
            ssl_setting = True
        elif "http" in profile_host:
            profile_host = profile_host[7:]
        if profile.port == 443:
            ssl_setting = True
        influx_host = "{}:{}".format(profile_host, profile.port)

        # TODO: remove hard-coded database credentials, should probably be read from a settings file
        if influx_host in influx_cred_map:
            (username, password) = influx_cred_map[influx_host]
        else:
            username = None
            password = None

        conn_settings = ConnectionSettings(
            host=profile.host,
            port=profile.port,
            username=username,
            password=password,
            database=profile.database,
            ssl=ssl_setting,
            verify_ssl=ssl_setting,
        )
        time_series_data = InfluxDBProfileManager(conn_settings)

        time_series_data.load_influxdb(
            profile.measurement,
            [profile.field],
            profile.startDate,
            profile.endDate,
        )

        for x in time_series_data.profile_data_list:
            if len(x) != 2:
                raise RuntimeError(
                    "InfluxDB profile currently only supports parsing exactly one "
                    "profile for each asset"
                )

        if not time_series_data.profile_data_list[0][0].tzinfo:
            index = pd.DatetimeIndex(
                data=[x[0] for x in time_series_data.profile_data_list],
                tz=datetime.timezone.utc,
            )
            logger.warning("No timezone specified for the input profile: default UTC has been used")
        else:
            index = pd.DatetimeIndex(data=[x[0] for x in time_series_data.profile_data_list])

        data = [x[1] for x in time_series_data.profile_data_list]
        series = pd.Series(data=data, index=index)
        self._df[profile.id] = series

        return series

    @staticmethod
    def _check_profile_time_series(
        profile_time_series: pd.Series, profile: esdl.InfluxDBProfile
    ) -> None:
        """
        Function that checks if the loaded profile matches what was expected

        Parameters
        ----------
        profile_time_series : the pandas Series of the profile obtained for the profile.
        profile : the InfluxDBProfile used to obtain the time series

        Returns
        -------
        None
        """
        if profile_time_series.index[0] != profile.startDate:
            raise RuntimeError(
                f"The user input profile start datetime: {profile.startDate} does not match the"
                f" start date in the database: {profile_time_series.index[0]} for asset: "
                f"{profile.field}"
            )
        if profile_time_series.index[-1] != profile.endDate:
            raise RuntimeError(
                f"The user input profile end datetime: {profile.endDate} does not match the end"
                f" datetime in the database: {profile_time_series.index[-1]} for asset: "
                f"{profile.field}"
            )

        # Error check: ensure that the profile data has a time resolution of 3600s (1hour) as
        # expected
        for d1, d2 in zip(profile_time_series.index, profile_time_series.index[1:]):
            if d2 - d1 != pd.Timedelta(hours=1):
                raise RuntimeError(
                    f"The timestep for variable {profile.field} between {d1} and {d2} isn't "
                    f"exactly 1 hour"
                )

    def _convert_profile_to_correct_unit(
        self, profile_time_series: pd.Series, profile: esdl.InfluxDBProfile
    ) -> pd.Series:
        """
        Conversion function to change the values in the provided series to the correct unit

        Parameters
        ----------
        profile_time_series: the time series obtained for the provided profile.
        profile: the profile which was used to obtain the series.

        Returns
        -------
        A pandas Series with the same index as the provided profile_time_series and with all values
        converted to either Watt or Joules, depending on the quantity used in the profile.
        """
        profile_quantity_and_unit = self._get_profile_quantity_and_unit(profile=profile)
        if profile_quantity_and_unit.physicalQuantity == esdl.PhysicalQuantityEnum.POWER:
            target_unit = POWER_IN_W
        elif profile_quantity_and_unit.physicalQuantity == esdl.PhysicalQuantityEnum.ENERGY:
            target_unit = ENERGY_IN_J
        elif profile_quantity_and_unit.physicalQuantity == esdl.PhysicalQuantityEnum.COST:
            if not (
                profile_quantity_and_unit.unit == esdl.UnitEnum.EURO
                and profile_quantity_and_unit.perUnit == esdl.UnitEnum.WATTHOUR
            ):
                raise RuntimeError(
                    f"For price profiles currently only profiles "
                    f"specified in euros per watt-hour are accepted,"
                    f"{profile} doesn't follow this convention."
                )
            return profile_time_series
        else:
            raise RuntimeError(
                f"The user input profile currently only supports loading profiles containing "
                f"either power, energy values or euros per Wh, not "
                f"{profile_quantity_and_unit.physicalQuantity}."
            )
        # The vectorized method below is used instead of profile_time_series.apply(), due to a
        # computational cost reduction (order of 2000 times faster for a profile with 8760
        # timesteps)
        return profile_time_series * convert_to_unit(
            value=1.0, source_unit=profile_quantity_and_unit, target_unit=target_unit
        )

    @staticmethod
    def _get_profile_quantity_and_unit(profile: esdl.InfluxDBProfile):
        try:
            return profile.profileQuantityAndUnit.reference
        except AttributeError:
            return profile.profileQuantityAndUnit


class ProfileReaderFromFile(BaseProfileReader):
    def __init__(self, energy_system: esdl.EnergySystem, file_path: Path):
        super().__init__(energy_system=energy_system, file_path=file_path)

    def _load_profiles_from_source(
        self,
        energy_system_components: Dict[str, Set[str]],
        esdl_asset_id_to_name_map: Dict[str, str],
        carrier_properties: Dict[str, Dict],
        ensemble_size: int,
    ) -> None:
        if self._file_path.suffix == ".xml":
            logger.warning(
                "XML type loading currently does not support loading " "price profiles for carriers"
            )
            self._load_xml(
                energy_system_components=energy_system_components,
                esdl_asset_id_to_name_map=esdl_asset_id_to_name_map,
            )
        elif self._file_path.suffix == ".csv":
            self._load_csv(
                energy_system_components=energy_system_components,
                carrier_properties=carrier_properties,
                ensemble_size=ensemble_size,
            )
        else:
            raise _ProfileParserException(
                f"Unsupported profile file extension " f"{self._file_path.suffix}"
            )

    def _load_csv(
        self,
        energy_system_components: Dict[str, Set[str]],
        carrier_properties: Dict[str, Dict],
        ensemble_size: int,
    ) -> None:
        data = pd.read_csv(self._file_path)
        try:
            timeseries_import_times = [
                datetime.datetime.strptime(entry.replace("Z", ""), "%Y-%m-%d %H:%M:%S").replace(
                    tzinfo=datetime.timezone.utc
                )
                for entry in data["DateTime"].to_numpy()
            ]
        except ValueError:
            try:
                timeseries_import_times = [
                    datetime.datetime.strptime(entry.replace("Z", ""), "%Y-%m-%dT%H:%M:%S").replace(
                        tzinfo=datetime.timezone.utc
                    )
                    for entry in data["DateTime"].to_numpy()
                ]
            except ValueError:
                try:
                    timeseries_import_times = [
                        datetime.datetime.strptime(
                            entry.replace("Z", ""), "%d-%m-%Y %H:%M"
                        ).replace(tzinfo=datetime.timezone.utc)
                        for entry in data["DateTime"].to_numpy()
                    ]
                except ValueError:
                    raise _ProfileParserException("Date time string is not in a supported format")

        logger.warning("Timezone specification not supported yet: default UTC has been used")

        self._reference_datetimes = timeseries_import_times

        for ensemble_member in range(ensemble_size):
            for component_type, var_name in self.component_type_to_var_name_map.items():
                for component_name in energy_system_components.get(component_type, []):
                    try:
                        values = data[f"{component_name.replace(' ', '')}"].to_numpy()
                    except KeyError:
                        pass
                    else:
                        self._profiles[ensemble_member][component_name + var_name] = values
            for properties in carrier_properties.values():
                carrier_name = properties.get("name")
                try:
                    values = data[carrier_name].to_numpy()
                except KeyError:
                    pass
                else:
                    self._profiles[ensemble_member][
                        carrier_name + self.carrier_profile_var_name
                    ] = values

    def _load_xml(self, energy_system_components, esdl_asset_id_to_name_map):
        timeseries_import_basename = self._file_path.stem
        input_folder = self._file_path.parent

        try:
            data = rtctools.data.pi.Timeseries(
                _ESDLInputDataConfig(esdl_asset_id_to_name_map, energy_system_components),
                input_folder,
                timeseries_import_basename,
                binary=False,
                pi_validate_times=False,
            )
        except IOError:
            raise Exception(
                "ESDLMixin: {}.xml not found in {}.".format(
                    timeseries_import_basename, input_folder
                )
            )

        # Convert timeseries timestamps to seconds since t0 for internal use
        if not data.times[0].tzinfo:
            for ii in range(len(data.times)):
                data.times[ii] = data.times[ii].replace(tzinfo=datetime.timezone.utc)
            logger.warning("No timezone specified for the input profile: default UTC has been used")

        self._reference_datetimes = data.times

        # Offer input timeseries to IOMixin
        for ensemble_member in range(data.ensemble_size):
            for variable, values in data.items(ensemble_member):
                self._profiles[ensemble_member][variable] = values


class _ESDLInputDataConfig:
    ns: dict = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}

    def __init__(self, id_map: Dict[str, str], energy_system_components: Dict[str, Set[str]]):
        # TODO: change naming source and demand to heat_source and heat_demand throughout code
        self.__id_map: Dict[str, str] = id_map
        self._sources: Set = set(energy_system_components.get("heat_source", []))
        self._demands: Set = set(energy_system_components.get("heat_demand", []))
        self._electricity_sources: Set = set(energy_system_components.get("electricity_source", []))
        self._electricity_demands: Set = set(energy_system_components.get("electricity_demand", []))
        self._gas_sources: Set = set(energy_system_components.get("gas_source", []))
        self._gas_demands: Set = set(energy_system_components.get("gas_demand", []))

    def variable(self, pi_header):
        location_id = pi_header.find("pi:locationId", self.ns).text

        try:
            component_name = self.__id_map[location_id]
        except KeyError:
            parameter_id = pi_header.find("pi:parameterId", self.ns).text
            qualifiers = pi_header.findall("pi:qualifierId", self.ns)
            qualifier_ids = ":".join(q.text for q in qualifiers)
            return f"{location_id}:{parameter_id}:{qualifier_ids}"

        if component_name in self._demands:
            suffix = ".target_heat_demand"
        elif component_name in self._sources:
            suffix = ".maximum_heat_source"
        elif component_name in self._electricity_demands:
            suffix = ".target_electricity_demand"
        elif component_name in self._electricity_sources:
            suffix = ".maximum_electricity_source"
        elif component_name in self._gas_demands:
            suffix = ".target_gas_demand"
        elif component_name in self._gas_sources:
            suffix = ".maximum_gas_source"
        else:
            logger.warning(
                f"Could not identify '{component_name}' as either source or demand. "
                f"Using neutral suffix '.target_heat' for its milp timeseries."
            )
            suffix = ".target_heat"

        # Note that the qualifier id (if any specified) refers to the profile
        # element of the respective ESDL asset->in_port. For now we just
        # assume that only milp demand timeseries are set in the XML file.
        return f"{component_name}{suffix}"



from .compressor import Compressor
from .gas_demand import GasDemand
from .gas_node import GasNode
from .gas_pipe import GasPipe
from .gas_source import GasSource
from .gas_substation import GasSubstation

__all__ = [
    "Compressor",
    "GasNode",
    "GasPipe",
    "GasSource",
    "GasDemand",
    "GasSubstation",
]
